# -*- coding: utf-8 -*-
"""Pro Custodibus multifactor authentication tool.

Checks WireGuard multifactor authentication state,
and enables the interactive portion of multifactor authentication.

Usage:
  procustodibus-mfa [--config=CONFIG] [-v | -vv | --verbosity=LEVEL]
  procustodibus-mfa --auth=USER --endpoint=ENDPOINT [--password-fd=NUMBER]
          [--secondary-code=CODE | --secondary-fd=NUMBER | --secondary-prompt]
          [--config=CONFIG] [-v | -vv | --verbosity=LEVEL]
  procustodibus-mfa --check --endpoint=ENDPOINT [--config=CONFIG]
  procustodibus-mfa --help
  procustodibus-mfa --version

Options:
  -h --help               Show this help
  --version               Show agent version
  -a --auth=USER          ID of user to authenticate as
  --password-fd=NUMBER    File descriptor to use instead of password prompt
  --secondary-code=CODE   Secondary verification code
  --secondary-fd=NUMBER   File descriptor to use for secondary code
  --secondary-prompt      Prompt for secondary code
  --check                 Check a single endpoint
  -e --endpoint=ENDPOINT  ID of endpoint to check/authenticate
  -c --config=CONFIG      Config file
  --verbosity=LEVEL       Log level (ERROR, WARNING, INFO, DEBUG)
  -v                      INFO verbosity
  -vv                     DEBUG verbosity

Examples:
  $ procustodibus-mfa
  Interface    Endpoint       ID           State
  -----------  -------------  -----------  ------------------------------
  wg-prod      Prod Server 1  ZSp2TSWJ6Ge  EXPIRED
  wg-prod      Prod Server 2  63Wa7kQ68oH  OK until 01/01/20 01:00:00 PST
  wg-dev       Test Server    USBiuu8PBZw  -
  $ procustodibus-mfa --check --endpoint=ZSp2TSWJ6Ge
  EXPIRED
  $ procustodibus-mfa --auth=Ahg1opVcGX --endpoint=ZSp2TSWJ6Ge
  Password: ********
  OK until 01/01/20 08:00:00 PST
"""
from docopt import docopt

from procustodibus_agent import __version__ as version
from procustodibus_agent.cnf import Cnf
from procustodibus_agent.mfa import check_mfa, do_mfa, list_mfa, read_password


def main():
    """Tool entry point."""
    args = docopt(__doc__)
    if args["--version"]:
        # print version to stdout
        print("procustodibus-mfa " + version)  # noqa: T201
    elif args["--auth"]:
        authenticate(
            args["--endpoint"],
            args["--auth"],
            args["--password-fd"],
            args["--secondary-code"],
            args["--secondary-fd"],
            args["--secondary-prompt"],
            args["--config"],
            args["--verbosity"] or args["-v"],
        )
    elif args["--check"]:
        check(args["--endpoint"], args["--config"])
    else:
        run(args["--config"], args["--verbosity"] or args["-v"])


# simpler to keep this logic together rather than subdivide it more functions
def authenticate(  # noqa: CFQ002
    endpoint,
    user,
    password_input=None,
    secondary_code=None,
    secondary_input=None,
    secondary_prompt=False,
    cnf_file="",
    verbosity=None,
):
    """Authenticates the specified user for the specified endpoint.

    Arguments:
        endpoint (str): Endpoint ID.
        user (str): User ID.
        password_input (str): Optional file descriptor from which to read password.
        secondary_code (str): Optional secondary verification code.
        secondary_input (str): Optional file descriptor from which to read code.
        secondary_prompt (bool): True to prompt for code.
        cnf_file (str): Optional path to config file.
        verbosity: Root log level.
    """
    if secondary_prompt:
        secondary_prompt = "Secondary code: "
    cnf = Cnf(cnf_file, verbosity)
    password = read_password(None, password_input, True)
    secondary_code = read_password(secondary_code, secondary_input, secondary_prompt)
    do_mfa(cnf, endpoint, user, password, secondary_code)


def check(endpoint, cnf_file=""):
    """Checks the MFA state of the specified endpoint.

    Arguments:
        endpoint (str): Endpoint ID.
        cnf_file (str): Path to config file (optional).
    """
    cnf = Cnf(cnf_file)
    check_mfa(cnf, endpoint)


def run(*args):
    """Lists MFA state for all endpoints of the configured host.

    Arguments:
        *args (list): List of arguments to pass to Cnf constructor.
    """
    cnf = Cnf(*args)
    list_mfa(cnf)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""Config utilities."""

import logging
import logging.config
import os
import sys
from contextlib import contextmanager
from os import environ
from pathlib import Path
from re import IGNORECASE
from re import compile as compile_re
from re import fullmatch, match, split, sub

from inflection import underscore

from procustodibus_agent import DEFAULT_API_URL, DEFAULT_APP_URL
from procustodibus_agent.windows.cnf import open_dpapi

DEFAULT_WG_CNF_DIRS = [
    "C:/Program Files/WireGuard/Data/Configurations",
    "/opt/homebrew/etc/wireguard",
    "/usr/local/etc/wireguard",
    "/etc/wireguard",
]
DEFAULT_CNF_DIRS = [
    ".",
    "C:/Program Files/Pro Custodibus Agent/cnf",
    "/usr/local/etc/procustodibus",
    "/etc/procustodibus",
] + DEFAULT_WG_CNF_DIRS
DEFAULT_DNS_SERVERS = ["9.9.9.9", "149.112.112.112", "2620:fe::fe", "2620:fe::9"]
DEFAULT_LOGGING_FORMAT = "%(asctime)s %(name)s %(levelname)s: %(message)s"
DEFAULT_LOGGING_LEVEL = "WARNING"
DEFAULT_SCRIPTS_DIRS = [
    "C:/Program Files/Pro Custodibus Agent/scripts",
    "/usr/local/lib/procustodibus/agent/scripts",
]
PROCUSTODIBUS_SPLITTABLE = {
    "dns": ",",
    "unmanaged_interfaces": ",",
}
WIREGUARD_SPLITTABLE = {
    "address": ",",
    "allowed_apps": ",",
    "allowed_ips": ",",
    "disallowed_apps": ",",
    "disallowed_ips": ",",
    "dns": ",",
    "post_down": ";",
    "post_up": ";",
    "pre_down": ";",
    "pre_up": ";",
}


def get_environ(name, default=None):
    """Looks up the specified environment variable.

    Arguments:
        name (str): Environment variable name (eg 'CONF').
        default: Default value (optional; defaults to None).

    Returns:
        Environment variable value or default.
    """
    return environ.get(f"PROCUSTODIBUS_{name}", default)


def init_log(root_level="", cnf=None):
    """Initializes python logging.

    Arguments:
        root_level (str): Root logging level. Defaults to 'WARNING'.
        cnf (str): Path to Python logging configuration file.
    """
    root_level = _get_root_level(root_level)

    if cnf and os.path.exists(cnf):
        logging.config.fileConfig(cnf)
        if root_level:
            logging.getLogger().setLevel(root_level)
        logging.getLogger(__name__).debug("init logging configuration from %s", cnf)
    else:
        _setup_default_logging(root_level)
        if root_level:
            logging.getLogger(__name__).debug("init logging at %s", root_level)


def _get_root_level(log_level):
    """Returns the log level value for the specified level or environment varable.

    Arguments:
        log_level (str): Log level.

    Returns:
        Log level.
    """
    level = get_environ("LOGGING_LEVEL") or log_level
    if level:
        level = level.upper()
    if level == "ALL":
        level = 1
    return level


def _setup_default_logging(log_level):
    """Sets up logging if no logging config file specified.

    Arguments:
        log_level (str): Log level.
    """
    log_format = get_environ("LOGGING_FORMAT") or DEFAULT_LOGGING_FORMAT
    log_level = log_level or DEFAULT_LOGGING_LEVEL
    logging.basicConfig(format=log_format, level=log_level, stream=sys.stdout)


@contextmanager
def open_ini(path, method="r"):
    """Opens the specified ini file for reading or writing (handling DPAPI encryption).

    Arguments:
        path (str): Path to file.
        method (str): "r" or "w" to read or write file (default "r").

    Yields:
        Stream object to read or write.
    """
    o = open_dpapi if str(path).endswith(".dpapi") else open
    with o(path, method) as f:
        yield f


def rename_ini(old_path, new_path):
    """Renames the specified ini file (handling DPAPI encryption).

    Arguments:
        old_path (str): Old path to file.
        new_path (str): New path to file.
    """
    old_path = Path(old_path)
    if old_path.suffix == ".dpapi":
        with open_dpapi(old_path, "r") as r:
            with open_dpapi(new_path, "w") as w:
                w.write(r.read())
        old_path.unlink()
    else:
        old_path.rename(new_path)


def find_default_cnf():
    """Finds the path to an existing config file.

    Returns:
        str: Path to an existing config file, or blank.
    """
    path = get_environ("CONF")
    if path:
        return path

    for directory in DEFAULT_CNF_DIRS:
        path = os.path.join(directory, "procustodibus.conf")
        if os.path.exists(path):
            return path

    return ""


def load_cnf(cnf_file):
    """Loads the specified procustodibus config file.

    Normalizes keys and values.

    Arguments:
        cnf_file (str): Path to config file, or blank.

    Returns:
        dict: Loaded config settings (dict of strings to normalized values).
    """
    return ini_to_dict(load_ini(cnf_file, PROCUSTODIBUS_SPLITTABLE))


def load_ini(ini_file, splittable=None):
    """Loads the specified ini config file.

    Converts section and key names to snake_case; splits and trims values.

    Arguments:
        ini_file (str): Path to config file, or blank.
        splittable (dict): Optional dict of key names whose values to split.

    Returns:
        dict: Loaded ini settings (dict of strings to list of dicts).
    """
    result = {}
    if not ini_file:
        return result

    section = {}
    _add_ini_value(result, "preface", section)
    if not splittable:
        splittable = {}

    with open_ini(ini_file) as f:
        for line in f:
            line = sub("#.*", "", line).strip()
            section_match = fullmatch(r"\s*\[([^\]]+)\]\s*", line)
            if section_match:
                section_name = underscore(section_match[1])
                section = {}
                _add_ini_value(result, section_name, section)
            elif line:
                _add_ini_line(section, line, splittable)

    return result


def _add_ini_line(section, line, splittable):
    """Parses the specified key=value line and adds it to the specified dict.

    Trims key and value, and converts key to snake_case.

    Splits value by splitter char if key found in splittable dict.

    Arguments:
        section (dict): Dict to add parsed line to.
        line (str): Line to parse (eg 'DNS = 10.0.0.1, wg').
        splittable (dict): Dict of snake_case keys to splitter characters.
    """
    parts = line.split("=", maxsplit=1)
    if len(parts) < 2:
        parts.append("true")

    # snake_case keys
    # with special-case for IP to convert AllowedIPs to allowed_ips
    key = underscore(sub("IP", "Ip", parts[0].strip()))
    splitter = splittable.get(key)
    if splitter:
        value = [x.strip() for x in parts[1].split(splitter)]
    else:
        value = parts[1].strip()

    _add_ini_value(section, key, value)


def _add_ini_value(container, key, value):
    """Adds specified ini value to list of values in specified dict.

    Arguments:
        container (dict): Dict to add value to.
        key (str): Key under which to add value.
        value: Value to add (string or list of strings).
    """
    existing = container.setdefault(key, [])
    if isinstance(value, list):
        existing.extend(value)
    else:
        existing.append(value)


# simpler to keep logic in same function
# even if it makes cognitive-complexity high
def load_ini_lines(ini_file):  # noqa: CCR001
    """Loads the specified ini config file as a list of lines in a list of sections.

    Includes comments, whitespace, and section headers as-is in lines. The first
    section (index 0) may be a "preface" section, before the first named section.

    Arguments:
        ini_file (str): Path to config file, or blank.

    Returns:
        list: List of list of lines.
    """
    if not ini_file:
        return []

    section = []
    result = [section]

    with open_ini(ini_file) as f:
        for line in f:
            section_match = match(r"\s*\[([^\]]+)\]", line)
            if section_match:
                section = _init_next_loaded_ini_section(result)
            section.append(line.rstrip())

    return [x for x in result if x]


def _init_next_loaded_ini_section(sections):
    """Initializes the next section and adds it to the existing list of sections.

    The main purpose of this fn is to find any comments from the last existing
    section that look like they belong to the next section; if found, it removes
    them from the last section, and adds them to the new section.

    Eg removes the last two lines from the following section, and adds them to
    the new section:
    ```
    # this peer
    # is 123
    [Peer]
    PublicKey = 123

    # next peer
    # is 456
    ```

    Arguments:
        sections (list): Existing list of sections.

    Returns:
        list: New section.
    """
    section = []
    if sections:
        prev_section = sections[-1]
        top_comment = _find_index_of_top_trailing_comment(prev_section)
        if top_comment < len(prev_section):
            section.extend(prev_section[top_comment:])
            del prev_section[top_comment:]
    sections.append(section)
    return section


def _find_index_of_top_trailing_comment(section):
    """Finds the index of trailing top comment in the specified list of lines.

    Eg returns an index of 5 for the following section lines:
    ```
    # this peer
    # is 123
    [Peer]
    PublicKey = 123

    # next peer
    # is 456
    ```

    Arguments:
        section (list): List of lines in a section.

    Returns:
        int: Index of top trailing comment, or zero.
    """
    for i, line in enumerate(reversed(section)):
        if not match(r"^# [^=]+", line):
            return len(section) - i
    return 0


# splitting up this function would make the logic harder to follow
# even if it makes cognitive-complexity high
def join_ini_splittable_lines(sections, splittable):  # noqa: CCR001
    """Updates the lines in the specified sections to join multi-line values.

    For example, given the following sections:
    ```
    [
        [
            '[Interface]',
            'Address = 10.0.0.1',
            'Address = fd10::1',
            'ListenPort = 51820',
            'PostUp = echo post',
            'PostUp = echo up',
            ''
        ],
    ]
    ```

    This would alter their lines to join lines with the same keys:
    ```
    [
        [
            '[Interface]',
            'Address = 10.0.0.1, fd10::1',
            'ListenPort = 51820',
            'PostUp = echo post; echo up',
            ''
        ],
    ]
    ```

    Arguments:
        sections (list): List of list of lines.
        splittable (dict): Dict of snake_case keys to join characters.
    """
    for lines in sections:
        previous_key = ""
        i = 0
        while i < len(lines):
            parts = lines[i].split("=", maxsplit=1)
            if len(parts) == 2:
                key = parts[0].strip()
                if previous_key == key and i > 0:
                    splitter = splittable.get(underscore(key), ",")
                    lines[i - 1] = f"{lines[i - 1]}{splitter}{parts[1]}"
                    del lines[i]
                    i -= 1
                else:
                    previous_key = key
            i += 1


def save_ini_lines(ini_file, sections, chmod=None):
    """Saves the specified ini config file from a list of lines in a list of sections.

    Arguments:
        ini_file (str): Path to config file.
        sections (list): List of list of lines.
        chmod (int): Optional octal mode value to set on file.
    """
    with open_ini(ini_file, "w") as f:
        for lines in sections:
            for line in lines:
                # print to file
                print(line, file=f)  # noqa: T201

    if chmod:
        ini_file.chmod(chmod)


def find_ini_section_with_line(sections, line):
    """Finds the specified section of lines containing the specified line start.

    Arguments:
        sections (list): List of list of lines.
        line (str): Line start to find.

    Returns:
        list: List of lines or None.
    """
    for section in sections:
        for x in section:
            if x.startswith(line):
                return section
    return None


# splitting up this function would make the logic harder to follow
# even if it makes cognitive-complexity high
def replace_ini_line_value(lines, key, value):  # noqa: CCR001
    """Replaces the specified key in the specified list with the specified value.

    Replaces all lines with the matched key; or if the key is not found, appends
    to the list. If the value is itself a list, adds a separate line for each value.

    If the value is None or an empty list ([]), simply removes all lines with
    the matched key.

    Arguments:
        lines (list): List of lines.
        key (str): Key to match (eg 'AllowedIPs').
        value: Value to replace (eg ['0.0.0.0/0', '::/0']).
    """
    re = compile_re(rf"\s*{key.lower()}\s*=", IGNORECASE)
    empty = value is None or value == []
    found = False

    i = 0
    while i < len(lines):
        if match(re, lines[i]):
            if found or empty:
                del lines[i]
                i -= 1
            else:
                found = True
                if isinstance(value, list):
                    lines[i] = f"{key} = {value[0]}"
                    for v in value[1:]:
                        i += 1
                        lines.insert(i, f"{key} = {v}")
                else:
                    lines[i] = f"{key} = {value}"
        i += 1

    if not found and not empty:
        append_ini_line_value(lines, key, value)


def append_ini_line_value(lines, key, value):
    """Appends the specified key=value pair to the specified list.

    If the value is itself a list, adds a separate line for each value.

    Arguments:
        lines (list): List of lines.
        key (str): Key to append (eg 'AllowedIPs').
        value: Value to append (eg ['0.0.0.0/0', '::/0']).
    """
    i = len(lines)
    while i > 0 and fullmatch(r"#.*|\s*", lines[i - 1]):
        i -= 1

    if isinstance(value, list):
        for v in value:
            lines.insert(i, f"{key} = {v}")
            i += 1
    else:
        lines.insert(i, f"{key} = {value}")


def ini_to_dict(ini):
    """Normalizes property values in the specified dict.

    Drops empty sections, selects first value of non-known-list properties,
    coerces known boolean and number properties.

    Arguments:
        ini (dict): Dict to convert.

    Returns:
        dict: Loaded config settings (dict of strings to normalized values).
    """
    root = {}

    for section_name, section in ini.items():
        src = section[0]
        if src:
            segments = section_name.split(".")
            dst = _create_dict_by_path(root, segments)
            _normalize_cnf_section(src, dst)

    return root


def _normalize_cnf_section(src, dst):
    """Normalizes the property values of the specified config section.

    Copies normalized values from src dict to dst dict.

    Arguments:
        src (dict): Source config section.
        dst (dict): Destination config section.
    """
    for key, value in src.items():
        value = _normalize_cnf_property(key, value)

        segments = key.split(".")
        if len(segments) > 1:
            item_dst = _create_dict_by_path(dst, segments[:-1])
            item_dst[segments[-1]] = value
        else:
            dst[key] = value


def _create_dict_by_path(root, segments):
    """Creates a hierarchy of dictionaries according to the specified path.

    Arguments:
        root (dict): Root under which to create dictionaries.
        segments (list): List of path segments (eg ['foo', 'bar', 'baz']).

    Returns:
        dict: Leaf dictionary.
    """
    for segment in segments:
        root = root.setdefault(segment, {})
    return root


def apply_cnf(obj, cnf):
    """Applies the specified config settings to the specified object.

    Arguments:
        obj: Object to apply settings to.
        cnf (dict): Config settings to apply.
    """
    for key, value in cnf.items():
        if key == "procustodibus":
            apply_cnf(obj, value)
        else:
            setattr(obj, key, value)


# keep all special-casing of values together in same function
# even if it makes cognitive-complexity high
def _normalize_cnf_property(key, value):  # noqa: CCR001
    """Adjusts the specified config property value if necessary.

    Arguments:
        key (str): Property name.
        value: Property value.

    Returns:
        Normalized property value.
    """
    # coerce list property values
    if key in ["dns", "unmanaged_interfaces"]:
        if isinstance(value, str):
            return [x for x in split(" *, *", value) if x]
        return value

    # default to blank string, extract first list value if list
    if not value:
        value = ""
    elif isinstance(value, list):
        value = value[0] or ""

    # coerce boolean property values
    if key in ["read_only", "redact_secrets"]:
        if isinstance(value, str):
            return True if match("[1TYty]", value) else False

    # coerce integer property values
    if key in ["loop"]:
        if isinstance(value, str):
            return int(value)

    return value


def get_file_last_modified(cnf_file):
    """Returns the last modified time of the specified procustodibus config file.

    Arguments:
        cnf_file (str): Path to config file.

    Returns:
        float: Last modified time in seconds.
    """
    f = Path(cnf_file)
    return f.stat().st_mtime if f.is_file() else 0


def reload_cnf_if_modified(cnf, cli_args=None):
    """Reloads the cnf file if modified.

    Arguments:
        cnf (Cnf): Config object.
        cli_args (list): List of arguments to pass to Cnf constructor when reloading.

    Returns:
        (Cnf, bool): Tuple of config object (reloaded or existing), and a boolean
        indicated whether the config was reloaded or not.
    """
    if cnf.watch_cnf:
        f = cnf.cnf_file or find_default_cnf()
        if cnf.cnf_file_last_modified != get_file_last_modified(f):
            cnf = Cnf(*cli_args) if cli_args else Cnf()
            return cnf, True
    return cnf, False


class Cnf(dict):
    """Configuration object."""

    # simpler to keep logic in same function even if it makes cognitive-complexity high
    def __init__(self, cnf_file="", verbosity=None, loop=0):  # noqa: CCR001
        """Creates new configuration object.

        Arguments:
            cnf_file (str): Path to configuration file. Defaults to no file.
            verbosity: Root log level. Defaults to 'WARNING'.
                If the number 1 is specified, sets log level to 'INFO'.
                If the number 2 is specified, sets log level to 'DEBUG'.
            loop (int): Seconds to sleep before looping. Defaults to 0 (no loop).
        """
        self.api = DEFAULT_API_URL
        self.app = DEFAULT_APP_URL
        self.wg = "wg"
        self.wg_quick = "wg-quick"
        self.wg_cnf_dir = find_default_wg_cnf_dir()
        self.agent = ""
        self.host = ""
        self.credentials = ""
        self.setup = ""
        self.dns = DEFAULT_DNS_SERVERS
        self.dns_protocol = ""
        self.doh = "dns.quad9.net"
        self.fw_mark = ""
        self.logging = ""
        self.loop = int(loop or 0)
        self.manager = ""
        self.one_line_fields = False
        self.read_only = False
        self.redact_secrets = False
        self.resolve_hostnames = "auto"
        self.resolver = None
        self.scripts = ""
        self.transport = None
        self.unmanaged_interfaces = []

        if not cnf_file:
            cnf_file = find_default_cnf()
        self.cnf_file = cnf_file
        apply_cnf(self, load_cnf(cnf_file))
        self.cnf_file_last_modified = get_file_last_modified(cnf_file)
        self.watch_cnf = bool(get_environ("WATCH_CONF"))

        self.wiresock = "wiresock" in self.wg
        if self.wiresock:
            self.one_line_fields = True

        if verbosity == 1:
            verbosity = "INFO"
        if verbosity == 2:
            verbosity = "DEBUG"
        if not self.logging:
            self.logging = _locate_extra_cnf_file(cnf_file, "logging")
        init_log(verbosity, self.logging)

        if self.dns == ["off"]:
            self.dns = []
        self.socket_mark = self.fw_mark

        if not self.credentials:
            self.credentials = _locate_extra_cnf_file(cnf_file, "credentials")
        if not self.setup:
            self.setup = _locate_extra_cnf_file(cnf_file, "setup")
        if not self.manager:
            self.manager = _default_manager()
        if not self.scripts:
            self.scripts = _find_scripts_dir()

    def __str__(self):  # noqa: D105
        return self.cnf_file

    def __bool__(self):  # noqa: D105
        return True


def _locate_extra_cnf_file(cnf_file, extra_type):
    """Builds path to extra conf file (like the credentials file).

    Arguments:
        cnf_file (str): Path to standard cnf file (eg '/etc/wireguard/agent.conf').
        extra_type (str): Extra type (eg 'credentials').

    Returns:
        str: Path to cnf file (eg '/etc/wireguard/agent-credentials.conf').
    """
    if not cnf_file:
        return ""

    p = Path(cnf_file)
    return f"{p.parent}/{p.stem}-{extra_type}{p.suffix}"


def _default_manager():
    """Determines name of default interface manager.

    Returns:
        str: Manager name (eg 'systemd').
    """
    for x in ["/lib", "/usr/lib"]:
        if os.path.exists(f"{x}/systemd/system/wg-quick@.service"):
            return "systemd"
    return ""


def _find_scripts_dir():
    """Determines path to scripts directory.

    Returns:
        str: Scripts directory (eg '/usr/local/lib/procustodibus/agent/scripts').
    """
    for directory in DEFAULT_SCRIPTS_DIRS:
        if os.path.exists(directory):
            return directory
    return ""


def find_default_wg_cnf_dir():
    """Finds the default /etc/wireguard directory or equivalent.

    Returns:
        str: Path to an existing directory, or blank.
    """
    for directory in DEFAULT_WG_CNF_DIRS:
        if os.path.exists(directory):
            return directory
    return ""



# -*- coding: utf-8 -*-
"""Utilities for applying and annotating extras."""

from logging import getLogger
from pathlib import Path
from platform import system
from re import compile as compile_re
from re import match, search, sub

from inflection import underscore

PEER_EXTRAS = [
    "allowed_apps",
    "disallowed_apps",
    "socks5_proxy",
    "socks5_proxy_username",
    "socks5_proxy_password",
    "socks5_proxy_all_traffic",
]


def apply_extras_to_scripts(cnf, interface):
    """Applies the extras from the specified interface properties to scripts.

    For example, given the following interface properties:
    ```
    {'extras': {'masquerade': 'outbound'}, 'post_up': ['echo up']}
    ```

    This would apply them like so:
    ```
    {
        'pre_up': ['scripts/masquerade pre_up "%i" "outbound"'],
        'post_up': ['echo up', 'scripts/masquerade post_up "%i" "outbound"'],
        'pre_down': ['scripts/masquerade pre_down "%i" "outbound"'],
        'post_down': ['scripts/masquerade post_down "%i" "outbound"'],
    }
    ```

    Arguments:
        cnf (Config): Config object.
        interface (dict): Dict of interface properties.

    Returns:
        dict: Same dict with additional properties.
    """
    extras = []
    for name, value in interface.get("extras", {}).items():
        value = _convert_extra_script_value_to_string(name, value)
        extras.append(f"{name} {value}")

    if extras:
        interface["extras"] = extras
    return interface


def _convert_extra_script_value_to_string(name, value):
    """Converts the specified extra script value to a string.

    Arguments:
        name (str): Extra script name.
        value: Extra script value.

    Returns:
        str: Extra script value as a string.

    Raises:
        ValueError: If the extra value contains problematic characters.
    """
    if value is True:
        return "true"
    if not value and not type(value) in (int, float):
        return ""

    value = str(value)
    bad_character = search(r"[\x00-\x1f\"\\]", value)
    if bad_character:
        encoded = hex(bad_character.group(0).encode("utf-8")[0])
        raise ValueError(f"bad character in extras value: {name} ({encoded})")
    return value


# simpler to keep logic in same function even if it makes cognitive-complexity high
def rewrite_ini_section_to_extra_scripts(cnf, lines, dev):  # noqa: CCR001
    """Rewrites the specified ini lines from real scripts to fake extra properties.

    For example, given the following lines:
    ```
    [
        '[Interface]',
        'PrivateKey = 123...',
        'PreUp = scripts/masquerade pre_up "%i" "outbound"',
        'PostUp = scripts/masquerade post_up "%i" "outbound"',
        'PreDown = scripts/masquerade pre_down "%i" "outbound"',
        'PostDown = scripts/masquerade post_down "%i" "outbound"',
        'PostUp = echo up',
    ]
    ```

    This would rewrite them like so:
    ```
    [
        '[Interface]',
        'PrivateKey = 123...',
        'Extras = masquerade outbound',
        'PostUp = echo up',
    ]
    ```

    Arguments:
        cnf (Config): Config object.
        lines (dict): List of lines in '[Interface]` section.
        dev (str): Name of interface to update (eg 'wg0').

    Returns:
        list: Same lines list with, but with updated lines.
    """
    re = _get_script_re(cnf.scripts)
    re = compile_re(rf"((?:Pre|Post)(?:Up|Down))\s*=\s*{re.pattern}")
    i = 0
    while i < len(lines):
        parts = re.match(lines[i])
        if parts:
            key, name, value = parts.groups()
            if key == "PreUp":
                lines[i] = f"Extras = {name} {value}"
            else:
                del lines[i]
        else:
            i += 1
    return lines


# simpler to keep logic in same function even if it makes cognitive-complexity high
def rewrite_extra_scripts_to_ini_section(cnf, lines, dev):  # noqa: CCR001
    """Rewrites the specified ini lines from fake extra properties to real scripts.

    For example, given the following lines:
    ```
    [
        '[Interface]',
        'PrivateKey = 123...',
        'Extras = masquerade outbound',
        'Extras = foo bar',
        'PostUp = echo up',
    ]
    ```

    This would rewrite them like so:
    ```
    [
        '[Interface]',
        'PrivateKey = 123...',
        'PreUp = scripts/masquerade pre_up "%i" "outbound"',
        'PostUp = scripts/masquerade post_up "%i" "outbound"',
        'PreDown = scripts/masquerade pre_down "%i" "outbound"',
        'PostDown = scripts/masquerade post_down "%i" "outbound"',
        'PostUp = echo up',
    ]
    ```

    Arguments:
        cnf (Config): Config object.
        lines (dict): List of lines in '[Interface]` section.
        dev (str): Name of interface to update (eg 'wg0').

    Returns:
        list: Same lines list with, but with updated lines.
    """
    i = 0
    while i < len(lines):
        line = lines[i]
        if not line.startswith("Extras ="):
            i += 1
            continue

        parts = match(r"Extras = (\w+) (.+)", line)
        if not parts:
            del lines[i]
            continue

        name, value = parts.groups()
        path = _get_script_path(cnf.scripts, name)
        if path:
            scripts = list(_expand_scripts(dev, path, name, value))
            lines[i : i + 1] = scripts
            i += len(scripts)
        else:
            getLogger(__name__).warning("no extras script: %s", name)
            del lines[i]
    return lines


def _get_script_path(directory, name):
    """Returns the path to the script if it exists, or "".

    Arguments:
        directory (str): Path to scripts directory (eg '/scripts').
        name (str): Name of script (eg 'forward').

    Returns:
        str: Path to script (eg '/scripts/forward').
    """
    if not directory:
        return ""

    windows = system() == "Windows"
    if windows:
        name = f"{name}.ps1"

    path = Path(directory, name)
    if path.is_file():
        return f'powershell -F "{path}"' if windows else str(path)
    return ""


def _expand_scripts(dev, path, name, value):
    """Creates the lines for an extra script.

    Arguments:
        dev (str): Name of interface to update (eg 'wg0').
        path (str): Full path to script (eg '/scripts/forward').
        name (str): Extra name (eg 'forward').
        value (str): Extra value (eg 'outbound').

    Yields:
        str: Expanded script line.
    """
    if system() != "Windows":
        dev = "%i"
    for phase in ["PreUp", "PostUp", "PreDown", "PostDown"]:
        yield f'{phase} = {path} {underscore(phase)} "{dev}" "{value}"'


def annotate_extras_from_scripts(cnf, interface):
    """Annotates the specified interface properties with extras from its scripts.

    For example, given the following interface properties:
    ```
    {
        'pre_up': ['scripts/masquerade pre_up "%i" "outbound"'],
        'post_up': ['echo up', 'scripts/masquerade post_up "%i" "outbound"'],
        'pre_down': ['scripts/masquerade pre_down "%i" "outbound"'],
        'post_down': ['scripts/masquerade post_down "%i" "outbound"'],
    }
    ```

    This would annotate it like so:
    ```
    {'extras': {'masquerade': 'outbound'}, 'post_up': ['echo up']}
    ```


    Arguments:
        cnf (Config): Config object.
        interface (dict): Dict of interface properties.

    Returns:
        dict: Same dict with additional properties.
    """
    directory = cnf.scripts
    if not directory:
        return interface

    extras = dict(_extract_extras_from_scripts(directory, interface))
    if not extras:
        return interface

    interface["extras"] = extras
    return _strip_extras_from_scripts(directory, interface)


def _extract_extras_from_scripts(directory, properties):
    """Extracts extra script names and values from the specified interface properties.

    Arguments:
        directory (str): Path to scripts directory (eg '/scripts').
        properties (dict): Dict of interface properties.

    Yields:
        tuple: Extra name and value (eg ('forward', 'outbound')).
    """
    re = _get_script_re(directory)
    for script in properties.get("pre_up", []):
        match = re.fullmatch(script)
        if match:
            yield match.groups()


def _strip_extras_from_scripts(directory, properties):
    """Strips extra scripts from the specified interface properties.

    Arguments:
        directory (str): Path to scripts directory (eg '/scripts').
        properties (dict): Dict of interface properties.

    Returns:
        dict: Same dict stripped of some values.
    """
    for phase in ["pre_up", "post_up", "pre_down", "post_down"]:
        _strip_extras_from_script(directory, properties, phase)
    return properties


def _strip_extras_from_script(directory, properties, phase):
    """Strips extra scripts from the specified script phase.

    Arguments:
        directory (str): Path to scripts directory (eg '/scripts').
        properties (dict): Dict of interface properties.
        phase (str): Pre/post/up/down phase (eg 'pre_up').

    Returns:
        dict: Same dict stripped of some values.
    """
    old_values = properties.get(phase)
    if not old_values:
        return properties

    re = _get_script_re(directory)
    new_values = [x for x in old_values if not re.fullmatch(x)]
    if not new_values:
        del properties[phase]
    elif len(old_values) != len(new_values):
        properties[phase] = new_values
    return properties


def _get_script_re(directory):
    """Generates regex to match name and value of extra scripts.

    Arguments:
        directory (str): Path to scripts directory (eg '/scripts').

    Returns:
        re: Regex that matches extra scripts.
    """
    if system() != "Windows":
        return compile_re(rf'{directory}\W(\w+) \S+ \S+ "([^"]*)"')
    directory = sub(r"[/\\]", r"\\W", directory)
    return compile_re(rf'powershell -F "{directory}\W(\w+).ps1" \S+ \S+ "([^"]*)"')


def apply_extras_to_peers(cnf, peers):
    """Applies the extras from the specified peer properties.

    For example, given the following peer properties:
    ```
    {'extras': {'allowed_apps': ['firefox', 'chrome']}}
    ```

    This would apply them like so:
    ```
    {'allowed_apps': ['firefox', 'chrome']}
    ```

    Arguments:
        cnf (Config): Config object.
        peers (list): List of peer dicts.

    Returns:
        list: Same peer dicts with additional properties.
    """
    if peers:
        for peer in peers:
            for name, value in peer.pop("extras", {}).items():
                peer[name] = value
    return peers


def annotate_extras_from_peers(cnf, interface):
    """Annotates the specified interface properties with extras from its peers.

    For example, given the following interface properties:
    ```
    {
        peers: {
            'ABC=': {'allowed_apps': ['firefox', 'chrome']},
        },
    }
    ```

    This would annotate it like so:
    ```
    {
        peers: [
            'ABC=': {'extras': {'allowed_apps': ['firefox', 'chrome']}},
        ],
    }
    ```

    Arguments:
        cnf (Config): Config object.
        interface (dict): Dict of interface properties.

    Returns:
        dict: Same dict with changes to peers.
    """
    peers = interface["peers"]
    if peers:
        for peer in peers.values():
            extras = {k: (peer.pop(k)) for k in PEER_EXTRAS if k in peer}
            if extras:
                peer["extras"] = extras
    return interface



# -*- coding: utf-8 -*-
"""API utilities."""

import os
import time
from base64 import urlsafe_b64encode
from datetime import datetime, timezone
from json import dumps
from random import randint  # noqa: DUO102
from socket import AI_CANONNAME, gaierror, getaddrinfo, gethostname

import requests
from nacl.encoding import Base64Encoder
from nacl.signing import SigningKey
from requests import Request, Session

from procustodibus_agent import DOCS_URL
from procustodibus_agent import __version__ as version
from procustodibus_agent.cnf import load_cnf
from procustodibus_agent.resolve_hostname import ResolverHTTPAdapter, get_resolver

API_TIMEOUT = 16  # seconds
UTC = timezone.utc


def get_session_transport(cnf):
    """Finds or creates session object cached for the specified config.

    Arguments:
        cnf (Config): Config object.

    Returns:
        Session: Requests session object.
    """
    session = cnf.transport
    if not session:
        adapter = ResolverHTTPAdapter(
            resolver=get_resolver(cnf),
            socket_mark=cnf.socket_mark,
        )
        session = Session()
        session.mount("http://", adapter)
        session.mount("https://", adapter)
        cnf.transport = session
    return session


def send_to_api(cnf, request, raises=True):
    """Sends request to the API.

    Arguments:
        cnf (Config): Config object.
        request (Request): Request object to send.
        raises (bool): True to raise on error response (default True).

    Returns:
        Response: Response object.
    """
    session = get_session_transport(cnf)

    if not hasattr(request, "body"):
        request = request.prepare()
    response = session.send(request, timeout=API_TIMEOUT)

    if raises:
        response.raise_for_status()
    return response


def request_from_api(cnf, method, url, **kwargs):
    """Sends request to the API.

    Arguments:
        cnf (Config): Config object.
        method (str): Request method (eg "GET").
        url (str): Relative request URL (eg "hosts/123").
        **kwargs (dict): Dict of arguments to pass to Request constructor.

    Returns:
        Response: Response object.
    """
    if not cnf.api:
        _raise_setup_issue("Missing conf for API endpoint")
    raises = kwargs.pop("raises", True)
    request = Request(method, f"{cnf.api}/{url}", **kwargs)
    return send_to_api(cnf, request, raises)


def setup_api(cnf):
    """Generates and saves new credential for configured agent.

    Arguments:
        cnf (Config): Config object.
    """
    raise_unless_has_cnf(cnf)
    setup = load_setup(cnf)
    code = setup["code"]

    signing_key = SigningKey.generate()
    public_key = signing_key.verify_key.encode(Base64Encoder).decode("utf-8")

    url = f"users/{cnf.agent}/credentials/signature"
    headers = {
        "authorization": f"X-Custos user=^{cnf.agent}, agent_setup=^{code}",
        "content-type": "application/json",
    }
    data = dumps(
        {
            "secret": public_key,
            "description": f"on {getfqdn()}",
            "keep_others": setup.get("keep_others") or False,
        },
        separators=(",", ":"),
    )

    request_from_api(cnf, "POST", url, data=data, headers=headers)
    save_signing_key(cnf, signing_key)


def login_api(cnf):
    """Logs into the api and saves session token on cnf obj.

    Arguments:
        cnf (Config): Config object.
    """
    raise_unless_has_cnf(cnf)

    challenge = get_signing_challenge(cnf)["data"][0]
    challenge_id = challenge["id"]
    signature = sign_data(cnf, challenge["attributes"]["value"].encode("utf-8"))

    headers = {
        "authorization": f"X-Custos user=^{cnf.agent}"
        f", challenge=^{challenge_id}"
        f", signature={signature}",
    }
    response = request_from_api(cnf, "POST", "sessions", headers=headers)

    # add 1-10 minutes of jitter before real session timeout
    # to prevent repeating stampeding herd on every login
    jitter = randint(60, 600)  # noqa: S311

    session = response.json()["data"][0]
    cnf.session = {
        "token": session["attributes"]["token"],
        "expiration": time.time() + session["meta"]["expiration_timeout"] - jitter,
    }


def ping_api(cnf, interfaces, executed=None):
    """Sends ping request to API.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Interface info parsed from `wg show`.
        executed (list): List of executed desired changes.

    Returns:
        Response: Response json.
    """
    raise_unless_has_cnf(cnf)

    data = {"interfaces": interfaces}
    if executed:
        data["executed"] = executed

    agent = build_agent_info(cnf)
    if agent:
        data["agent"] = agent

    response = request_with_session(
        cnf,
        "POST",
        f"hosts/{cnf.host}/ping/{version}",
        data=dumps(data, separators=(",", ":")),
        headers={"content-type": "application/json"},
    )
    return response.json()


def get_health_info(cnf):
    """Gets the api's health info.

    Arguments:
        cnf (Config): Config object.

    Returns:
        Response: Response json.
    """
    return request_from_api(cnf, "GET", "health").json()


def get_host_info(cnf):
    """Gets the configured host's info from the API.

    Arguments:
        cnf (Config): Config object.

    Returns:
        Response: Response json.
    """
    return request_with_session(cnf, "GET", f"hosts/{cnf.host}").json()


def send_with_session(cnf, request):
    """Sends request with session authn header.

    Arguments:
        cnf (Config): Config object.
        request (Request): Request object to send.

    Returns:
        Response: Response object.
    """
    create_session_if_expired(cnf)
    prepped = request.prepare()
    prepped.headers["authorization"] = build_authn_header(cnf)

    response = send_to_api(cnf, prepped, False)

    if response.status_code == requests.codes.unauthorized:
        login_api(cnf)
        prepped = request.prepare()
        prepped.headers["authorization"] = build_authn_header(cnf)
        response = send_to_api(cnf, prepped, False)

    response.raise_for_status()
    return response


def request_with_session(cnf, method, url, **kwargs):
    """Sends request with session authn header.

    Arguments:
        cnf (Config): Config object.
        method (str): Request method (eg "GET").
        url (str): Relative request URL (eg "hosts/123").
        **kwargs (dict): Dict of arguments to pass to Request constructor.

    Returns:
        Response: Response object.
    """
    request = Request(method, f"{cnf.api}/{url}", **kwargs)
    return send_with_session(cnf, request)


def create_session_if_expired(cnf):
    """Creates a new session if session is missing or has timed-out.

    Arguments:
        cnf (Config): Config object.
    """
    raise_unless_has_cnf(cnf)
    session = cnf.__dict__.get("session")
    if not session or session["expiration"] < time.time():
        login_api(cnf)


def build_authn_header(cnf):
    """Constructs authn header with session token.

    Arguments:
        cnf (Config): Config object.

    Returns:
        str: Authn header
    """
    return f"X-Custos user=^{cnf.agent}, session=^{cnf.session['token']}"


def get_signing_challenge(cnf):
    """Gets the api's current authn challenge data.

    Arguments:
        cnf (Config): Config object.

    Returns:
        Response: Response json.
    """
    return request_from_api(cnf, "GET", "sessions/challenge").json()


def sign_data(cnf, data):
    """Signs the specified message.

    Arguments:
        cnf (Config): Config object.
        data (bytes): Message to sign.

    Returns:
        str: Base64-web-encoded signature.
    """
    signing_key = load_signing_key(cnf)
    signed_message = signing_key.sign(data)
    return encode_base64_web(signed_message.signature)


def load_signing_key(cnf):
    """Loads the agent's signing key.

    Arguments:
        cnf (Config): Config object.

    Returns:
        SingingKey: Agent's signing key.

    Propagates:
        ValueError: If the signing credentials are missing or invalid.
    """
    try:
        credentials = _load_credentials_dict(cnf.credentials, "credentials")
        private_key = credentials["private_key"]
    except KeyError:
        _raise_setup_issue("Missing private key in credentials file")

    try:
        return SigningKey(private_key, Base64Encoder)
    except Exception:
        _raise_setup_issue("Invalid private key in credentials file")


def encode_base64_web(message):
    """Base64-web-encodes the specified message.

    Arguments:
        message (str): Message to encode.

    Returns:
        str: Base64-web-encoded message.
    """
    return urlsafe_b64encode(message).decode("utf-8").rstrip("=")


def format_datetime(now=None):
    """Formats datetime as an RFC 3339 string.

    Arguments:
        now (datetime): Datetime to format (defaults to now).

    Returns:
        str: Formatted datetime.
    """
    if not now:
        now = time.time()
    return datetime.fromtimestamp(now, tz=UTC).strftime("%Y-%m-%dT%H:%M:%SZ")


def build_agent_info(cnf):
    """Builds dict of agent info to send to API.

    Arguments:
        cnf (Config): Config object.

    Returns:
        dict: Agent info dict.
    """
    keys = ["read_only", "redact_secrets", "unmanaged_interfaces"]
    if cnf.resolve_hostnames != "auto":
        keys.append("resolve_hostnames")

    return {key: getattr(cnf, key) for key in keys if getattr(cnf, key)}


def load_setup(cnf):
    """Loads the setup configuration dict, or raises an error.

    Arguments:
        cnf (Config): Config object.

    Returns:
        dict: Setup configuration dict.

    Propagates:
        ValueError: If the setup config is missing the "code" property, or
                    the "expires" property indicates it has expired, or
                    if generated credentials cannot be saved because there is
                    no configured path to the credentials file.
    """
    if type(cnf.credentials) is dict:
        _raise_setup_issue("Missing path to credentials file")

    try:
        setup = _load_credentials_dict(cnf.setup, "setup")
        setup["code"]
    except KeyError:
        _raise_setup_issue("Missing code in setup file")

    expires = setup.get("expires")
    if expires:
        expires = datetime.strptime(expires, "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=UTC)
        now = datetime.now(tz=UTC)
        if expires < now:
            _raise_setup_issue("Setup code has expired")

    return setup


def save_signing_key(cnf, signing_key):
    """Saves the specified signing key to the configured credentials location.

    Also deletes the configured setup file (to indicate setup is complete).

    Arguments:
        cnf (Config): Config object.
        signing_key (SigningKey): Signing key to save.
    """
    private_key = signing_key.encode(Base64Encoder).decode("utf-8")
    public_key = signing_key.verify_key.encode(Base64Encoder).decode("utf-8")

    credentials = f"""
        # procustodibus-credentials.conf generated {format_datetime()}
        # for agent {cnf.agent} on {getfqdn()}
        [Procustodibus.Credentials]
        PublicKey = {public_key}
        PrivateKey = {private_key}
    """.strip()

    with open(cnf.credentials, "w") as f:
        for line in credentials.splitlines():
            # print to file
            print(line.strip(), file=f)  # noqa: T201

    os.chmod(cnf.credentials, 0o640)

    if type(cnf.setup) is not dict:
        os.remove(cnf.setup)


def getfqdn():
    """Looks up fully-qualified domain name of localhost.

    Returns:
        str: Fully-qualified domain name (eg 'foo.example.com').
    """
    hostname = gethostname()
    try:
        return getaddrinfo(hostname, 0, flags=AI_CANONNAME)[0][3]
    except gaierror:
        return hostname


def raise_unless_has_cnf(cnf):
    """Raises an error unless the specified cnf has all required settings.

    Arguments:
        cnf (Config): Config object.

    Propagates:
        ValueError: If the specified cnf is missing some required settings.
    """
    missing = []

    if not cnf.api:
        missing.append("API endpoint")

    if not cnf.agent:
        missing.append("Agent ID")

    if not cnf.host:
        missing.append("Host ID")

    if missing:
        _raise_setup_issue(f"Missing conf for {' and '.join(missing)}")


def _raise_setup_issue(problem):
    raise ValueError(f"{problem}; see {DOCS_URL}/guide/agents/troubleshoot/ to fix")


def _load_credentials_dict(path, key):
    """Loads the credentials in the specified ini file if not already a dict.

    Arguments:
        path: Path to credentials file or dict of credentials.
        key: Key to load in ini (eg 'credentials').

    Returns:
        dict: Dict containing credentials.

    Propagates:
        ValueError: If the specified file cannot be read.
    """
    if type(path) is dict:
        return path

    try:
        loaded = load_cnf(path)
    except OSError:
        _raise_setup_issue(f"Could not read {key} file {path}")

    return loaded["procustodibus"][key]



# -*- coding: utf-8 -*-
"""Pro Custodibus Agent."""

__version__ = "1.7.1"

DEFAULT_API_URL = "https://api.custodib.us"
DEFAULT_APP_URL = "https://pro.custodib.us"
DOCS_URL = "https://docs.procustodibus.com"



# -*- coding: utf-8 -*-
"""Utilities for manipulating wq conf ini."""

from contextlib import suppress
from logging import DEBUG, getLogger
from pathlib import Path
from platform import system

from inflection import underscore

from procustodibus_agent import __version__ as version
from procustodibus_agent.api import format_datetime
from procustodibus_agent.cnf import (
    WIREGUARD_SPLITTABLE,
    find_ini_section_with_line,
    join_ini_splittable_lines,
    load_ini,
    load_ini_lines,
    rename_ini,
    replace_ini_line_value,
    save_ini_lines,
)
from procustodibus_agent.extras import (
    annotate_extras_from_peers,
    annotate_extras_from_scripts,
    apply_extras_to_peers,
    apply_extras_to_scripts,
    rewrite_extra_scripts_to_ini_section,
    rewrite_ini_section_to_extra_scripts,
)
from procustodibus_agent.resolve_hostname import (
    resolve_endpoint_hostname,
    split_endpoint_hostname_and_port,
)
from procustodibus_agent.wg import derive_public_key, hash_preshared_key


def load_all_from_wg_cnf(cnf):
    """Loads equivalent of `wg show` from wg all configs.

    Arguments:
        cnf (Config): Config object.

    Returns:
        dict: Interfaces.
    """
    directory = Path(cnf.wg_cnf_dir)
    if not cnf.wg_cnf_dir or not directory.is_dir():
        return {}

    suffix = ".conf.dpapi" if system() == "Windows" else ".conf"
    interfaces = {}
    for f in directory.glob(f"*{suffix}"):
        name = f.name[: -len(suffix)]
        interface = load_interface_from_wg_cnf(cnf, name)
        if interface:
            interfaces[name] = interface
    return interfaces


def load_interface_from_wg_cnf(cnf, name):
    """Loads equivalent of `wg show` from specified wg config.

    Arguments:
        cnf (Config): Config object.
        name (str): Interface name (eg 'wg0').

    Returns:
        dict: Interface properties
    """
    properties = {}
    path = find_wg_cnf_path(cnf, name)
    try:
        if not path.exists():
            return properties
    except PermissionError:
        return properties

    ini = load_ini(path, WIREGUARD_SPLITTABLE)
    interface = ini.get("interface")
    if not interface:
        return properties

    _core_interface_properties(properties, interface[0])
    _derive_public_key(properties)
    _annotate_interface_properties(properties, interface[0])
    annotate_extras_from_scripts(cnf, properties)
    _core_interface_peers(cnf, properties, ini)
    annotate_extras_from_peers(cnf, properties)
    return properties


# keep conversion of all interface values together in same function
# even if it makes cognitive-complexity high
def _core_interface_properties(properties, ini):  # noqa: CCR001
    """Loads core interface properties from specified wg config.

    Arguments:
        properties (dict): Dict of interface properties.
        ini (dict): Dict of interface config from wg ini.

    Returns:
        dict: Same dict with additional properties.
    """
    # string properties
    for key in ["private_key"]:
        values = ini.get(key)
        if values:
            properties[key] = values[0]

    # integer properties
    for key, ini_key in {"fwmark": "fw_mark", "listen_port": "listen_port"}.items():
        values = ini.get(ini_key)
        if values:
            v = values[0]
            with suppress(ValueError):
                properties[key] = int(v, 16) if v.startswith("0x") else int(v)

    return properties


def _core_interface_peers(cnf, interface, ini):
    """Loads core interface peers from specified wg config.

    Arguments:
        cnf (Config): Config object.
        interface (dict): Dict of interface properties.
        ini (dict): Dict of full config from wg ini.

    Returns:
        dict: Same dict with additional properties.
    """
    peers = {}
    for peer in ini.get("peer", []):
        properties = _core_peer_properties(peer)
        public_key = properties.pop("public_key", None)
        if public_key:
            peers[public_key] = properties
            _wiresock_peer_properties(cnf, peer, properties)
            _hash_preshared_key(properties)
            _resolve_endpoint(cnf, properties)
    interface["peers"] = peers
    return interface


# keep conversion of all interface values together in same function
# even if it makes cognitive-complexity high
def _core_peer_properties(ini):  # noqa: CCR001
    """Loads core peer properties from specified wg config.

    Arguments:
        ini (dict): Dict of peer config from wg ini.

    Returns:
        dict: Peer properties
    """
    properties = {}

    # string properties
    for key in ["endpoint", "preshared_key", "public_key"]:
        values = ini.get(key)
        if values:
            properties[key] = values[0]

    # integer properties
    for key in ["persistent_keepalive"]:
        values = ini.get(key)
        if values:
            with suppress(ValueError):
                properties[key] = int(values[0])

    # list properties
    for key in ["allowed_ips"]:
        values = ini.get(key)
        if values:
            properties[key] = values

    return properties


# keep conversion of all interface values together in same function
# even if it makes cognitive-complexity high
def _wiresock_peer_properties(cnf, ini, properties):  # noqa: CCR001
    """Loads core peer properties from specified wg config.

    Arguments:
        cnf (Config): Config object.
        ini (dict): Dict of peer config from wg ini.
        properties (dict): Dict of peer properties.

    Returns:
        dict: Peer properties
    """
    if not cnf.wiresock:
        return properties

    # string properties
    for key in ["socks5_proxy", "socks5_proxy_username", "socks5_proxy_password"]:
        values = ini.get(key)
        if values:
            properties[key] = values[0]

    # boolean properties
    for key in ["socks5_proxy_all_traffic"]:
        values = ini.get(key)
        if values:
            properties[key] = values[0] == "true"

    # list properties
    for key in ["allowed_apps", "disallowed_apps", "disallowed_ips"]:
        values = ini.get(key)
        if values:
            properties[key] = values

    return properties


def _derive_public_key(properties):
    """Adjusts interface properties to derive public key.

    Arguments:
        properties (dict): Dict of interface properties.

    Returns:
        dict: Same dict with additional properties.
    """
    private_key = properties.get("private_key")
    if private_key:
        properties["public_key"] = derive_public_key(private_key)
    return properties


def _hash_preshared_key(properties):
    """Adjusts peer properties to hash preshared key.

    Arguments:
        properties (dict): Dict of peer properties.

    Returns:
        dict: Same dict with additional properties.
    """
    preshared_key = properties.get("preshared_key")
    if preshared_key:
        properties["preshared_key_hash"] = hash_preshared_key(preshared_key)
    return properties


def _resolve_endpoint(cnf, properties):
    """Adjusts peer properties to include numeric endpoint and separate hostname.

    Arguments:
        cnf (Config): Config object.
        properties (dict): Dict of peer properties.

    Returns:
        dict: Same dict with additional properties.
    """
    endpoint, hostname, _ = resolve_endpoint_hostname(cnf, properties.get("endpoint"))
    if hostname:
        properties["hostname"] = hostname
        properties["endpoint"] = endpoint
    return properties


def annotate_wg_show_with_wg_cnf(cnf, interfaces):
    """Annotates parsed output of `wg show` with wg config.

    Arguments:
        cnf (Config): Config object.
        interfaces (dict): Dict parsed from `wg show` command.

    Returns:
        dict: Same dict with additional properties.
    """
    for name, properties in interfaces.items():
        _annotate_interface(cnf, name, properties)
    return interfaces


def find_wg_cnf_path(cnf, name):
    """Returns path to ini file for specified wg interface.

    Arguments:
        cnf (Config): Config object.
        name (str): Interface name (eg 'wg0').

    Returns:
        Path: Path to ini file (eg '/etc/wireguard/wg0.conf').
    """
    directory = cnf.wg_cnf_dir or "/etc/wireguard"
    file_name = f"{name}.conf.dpapi" if system() == "Windows" else f"{name}.conf"
    return Path(directory, file_name)


def _annotate_interface(cnf, name, properties):
    """Annotates specified dict with wg config for specified interface.

    Arguments:
        cnf (Config): Config object.
        name (str): Interface name (eg 'wg0').
        properties (dict): Dict of interface properties.

    Returns:
        dict: Same dict with additional properties.
    """
    path = find_wg_cnf_path(cnf, name)
    try:
        if not path.exists():
            return properties
    except PermissionError:
        return properties

    ini = load_ini(path, WIREGUARD_SPLITTABLE)
    interface = ini.get("interface")
    if not interface:
        return properties

    _annotate_interface_properties(properties, interface[0])
    annotate_extras_from_scripts(cnf, properties)
    _annotate_interface_peers(properties, ini)
    annotate_extras_from_peers(cnf, properties)
    return properties


# keep conversion of all interface values together in same function
# even if it makes cognitive-complexity high
def _annotate_interface_properties(properties, ini):  # noqa: CCR001
    """Annotates specified interface properties from specified wg config.

    Arguments:
        properties (dict): Dict of interface properties.
        ini (dict): Dict of interface config from wg ini.

    Returns:
        dict: Same dict with additional properties.
    """
    # string properties
    for key in ["table"]:
        values = ini.get(key)
        if values:
            properties[key] = values[0]

    # integer properties
    for key in ["mtu"]:
        values = ini.get(key)
        if values:
            with suppress(ValueError):
                properties[key] = int(values[0])

    # boolean properties
    for key in ["save_config"]:
        values = ini.get(key)
        if values:
            properties[key] = values[0] == "true"

    # list properties
    for key in ["address", "dns", "pre_up", "post_up", "pre_down", "post_down"]:
        values = ini.get(key)
        if values:
            properties[key] = values

    return properties


def _annotate_interface_peers(interface, ini):
    """Annotates specified interface peers from specified wg config.

    Arguments:
        interface (dict): Dict of interface properties.
        ini (dict): Dict of full config from wg ini.

    Returns:
        dict: Same dict with additional properties.
    """
    peers = {p["public_key"][0]: p for p in ini.get("peer", []) if p.get("public_key")}
    for name, properties in interface.get("peers", {}).items():
        peer = peers.get(name)
        if peer:
            _annotate_peer_properties(properties, peer)
    return interface


def _annotate_peer_properties(properties, ini):
    """Annotates specified peer properties from specified wg config.

    Arguments:
        properties (dict): Dict of peer properties.
        ini (dict): Dict of peer config from wg ini.

    Returns:
        dict: Same dict with additional properties.
    """
    endpoint = ini.get("endpoint")
    if endpoint:
        hostname, _ = split_endpoint_hostname_and_port(endpoint[0])
        if hostname:
            properties["hostname"] = hostname

    return properties


def delete_wg_cnf(cnf, name):
    """Deletes the cnf file for the specified interface with the specified name.

    Arguments:
        cnf (Config): Config object.
        name (str): Interface name (eg 'wg0').
    """
    path = find_wg_cnf_path(cnf, name)
    if path.exists():
        path.unlink()


def rename_wg_cnf(cnf, old_name, new_name):
    """Renames the cnf file for the specified interface to the specified new name.

    Arguments:
        cnf (Config): Config object.
        old_name (str): Existing name (eg 'wg0').
        new_name (str): New name (eg 'wg100').
    """
    if old_name and new_name and old_name != new_name:
        path = find_wg_cnf_path(cnf, old_name)
        if path.exists():
            rename_ini(path, find_wg_cnf_path(cnf, new_name))


def update_wg_cnf(cnf, name, interface, peers=None):
    """Updates the cnf file for the specified interface with the specified properties.

    Arguments:
        cnf (Config): Config object.
        name (str): Name of interface to update (eg 'wg0').
        interface (dict): Properties of interface to update.
        peers (list): List of dicts with peer properties to update.

    Raises:
        ValueError: Cnf file for the interface cannot be written.
    """
    if not interface and not peers:
        return

    chmod = None
    path = find_wg_cnf_path(cnf, name)
    if path.exists():
        sections = load_ini_lines(path)
    elif interface.get("private_key"):
        sections = [_stub_wg_cnf_interface_section(cnf, interface)]
        chmod = 0o640
    else:
        raise ValueError(f"{path} does not exist and private key not available")

    interface = apply_extras_to_scripts(cnf, interface)
    peers = apply_extras_to_peers(cnf, peers)

    section = _find_wg_cnf_interface_section(cnf, interface, path, sections)
    rewrite_ini_section_to_extra_scripts(cnf, section, name)
    _update_wg_cnf_interface_section(section, interface)
    _update_wg_cnf_routing_section(section, interface)
    _update_wg_cnf_peer_sections(cnf, sections, peers)
    rewrite_extra_scripts_to_ini_section(cnf, section, name)

    if cnf.one_line_fields:
        join_ini_splittable_lines(sections, WIREGUARD_SPLITTABLE)
    _debug_wg_cnf_sections(sections)
    save_ini_lines(path, sections, chmod)


def _find_wg_cnf_interface_section(cnf, interface, path, sections):
    """Finds the interface section within the list of sections.

    Arguments:
        cnf (Config): Config object.
        interface (dict): Properties of interface to update.
        path (str): Path to cnf file.
        sections (list): List of list of lines.

    Returns:
        list: List of lines in interface section.

    Raises:
        ValueError: Interface section cannot be found or stubbed.
    """
    section = find_ini_section_with_line(sections, "[Interface]")
    if section:
        return section

    if interface.get("private_key"):
        section = _stub_wg_cnf_interface_section(cnf, interface)
        sections.append(section)
        return section
    else:
        raise ValueError(
            f"{path} does not contain an existing interface definition "
            "and private key not available"
        )


def _stub_wg_cnf_interface_section(cnf, interface):
    """Generates lines for the interface section of a new cnf file.

    Arguments:
        cnf (Config): Config object.
        interface (dict): Properties of interface.

    Returns:
        list: List of lines for new cnf file.
    """
    identifier = interface.get("id")
    description = interface.get("description")
    url = f"{cnf.app}/interfaces/{identifier}" if identifier else ""

    if description and url:
        description = f"{description} ({url})"
    else:
        description = description or url

    return [
        f"# generated {format_datetime()} by procustodibus-agent {version}",
        f"# {description}",
        "[Interface]",
        f"PrivateKey = {interface['private_key']}",
    ]


def _update_wg_cnf_interface_section(lines, interface):
    """Updates the lines of the interface section of a cnf file.

    Arguments:
        lines (list): List of lines in interface section.
        interface (dict): Core properties of interface to update.
    """
    pk = interface.get("private_key")
    if pk:
        replace_ini_line_value(lines, "PrivateKey", pk)

    _update_wg_cnf_number_line(lines, interface, "ListenPort", ["listen_port", "port"])
    _update_wg_cnf_number_line(lines, interface, "FwMark", ["fw_mark", "fwmark"])


def _update_wg_cnf_routing_section(lines, interface):
    """Updates the lines of the interface section of a cnf file with routing info.

    Arguments:
        lines (list): List of lines in interface section.
        interface (dict): Routing properties of interface to update.
    """
    _update_wg_cnf_list_line(lines, interface, "Address")

    dns = interface.get("dns")
    search = interface.get("search")
    if dns is not None and search is not None:
        dns = dns + search
    elif search is not None:
        dns = search
    if dns == []:
        replace_ini_line_value(lines, "DNS", None)
    elif dns:
        replace_ini_line_value(lines, "DNS", dns)

    _update_wg_cnf_number_line(lines, interface, "MTU")
    _update_wg_cnf_text_auto_line(lines, interface, "Table")

    for key in ["Extras", "PreUp", "PostUp", "PreDown", "PostDown"]:
        _update_wg_cnf_list_line(lines, interface, key)

    _update_wg_cnf_boolean_line(lines, interface, "SaveConfig")


# keep top-level logic for updating peer sections together
# even if it makes cognitive-complexity high
def _update_wg_cnf_peer_sections(cnf, sections, peers):  # noqa: CCR001
    """Updates the lines of the peer sections of a cnf file.

    Arguments:
        cnf (Config): Config object.
        sections (list): List of list of lines.
        peers (list): List of dicts with peer properties to update.
    """
    for peer in peers or []:
        pk = peer["public_key"]
        section = find_ini_section_with_line(sections, f"PublicKey = {pk}")
        if peer.get("delete"):
            if section:
                sections.remove(section)
        else:
            if not section:
                section = _stub_wg_cnf_peer_section(cnf, peer)
                sections.append(section)
            _update_wg_cnf_peer_section(cnf, section, peer)


def _stub_wg_cnf_peer_section(cnf, peer):
    """Generates lines for a new peer section of a cnf file.

    Arguments:
        cnf (Config): Config object.
        peer (dict): Properties of the peer.

    Returns:
        list: List of lines for new peer.
    """
    identifier = peer.get("id")
    name = peer.get("name")
    url = f"{cnf.app}/endpoints/{identifier}" if identifier else ""

    if name and url:
        description = f"{name} ({url})"
    else:
        description = name or url

    return [
        "",
        f"# {description}",
        "[Peer]",
        f"PublicKey = {peer['public_key']}",
    ]


def _update_wg_cnf_peer_section(cnf, lines, peer):
    """Updates the lines of a peer section in a cnf file.

    Arguments:
        cnf (Config): Config object.
        lines (list): Existing list of lines in the peer section.
        peer (dict): Properties of the peer to update.
    """
    _update_wg_cnf_text_off_line(lines, peer, "PresharedKey")
    _update_wg_cnf_list_line(lines, peer, "AllowedIPs", ["allowed_ips"])
    _update_wg_cnf_text_line(lines, peer, "Endpoint")
    _update_wg_cnf_number_line(
        lines, peer, "PersistentKeepalive", ["persistent_keepalive", "keepalive"]
    )

    if cnf.wiresock:
        _update_wg_cnf_list_line(lines, peer, "AllowedApps")
        _update_wg_cnf_list_line(lines, peer, "DisallowedApps")
        _update_wg_cnf_list_line(lines, peer, "DisallowedIPs", ["disallowed_ips"])
        _update_wg_cnf_text_line(lines, peer, "Socks5Proxy")
        _update_wg_cnf_text_line(lines, peer, "Socks5ProxyUsername")
        _update_wg_cnf_text_line(lines, peer, "Socks5ProxyPassword")
        _update_wg_cnf_boolean_line(lines, peer, "Socks5ProxyAllTraffic")


def _update_wg_cnf_list_line(lines, src, ini_key, src_keys=None):
    _update_wg_cnf_line(lines, src, [[]], ini_key, src_keys)


def _update_wg_cnf_text_line(lines, src, ini_key, src_keys=None):
    _update_wg_cnf_line(lines, src, [""], ini_key, src_keys)


def _update_wg_cnf_text_auto_line(lines, src, ini_key, src_keys=None):
    _update_wg_cnf_line(lines, src, ["", "auto"], ini_key, src_keys)


def _update_wg_cnf_text_off_line(lines, src, ini_key, src_keys=None):
    _update_wg_cnf_line(lines, src, ["", "off"], ini_key, src_keys)


def _update_wg_cnf_number_line(lines, src, ini_key, src_keys=None):
    _update_wg_cnf_line(lines, src, [0, "0", "off"], ini_key, src_keys)


def _update_wg_cnf_boolean_line(lines, src, ini_key, src_keys=None):
    _update_wg_cnf_line(lines, src, [False, "", "false"], ini_key, src_keys)


def _update_wg_cnf_line(lines, src, empty_values, ini_key, src_keys=None):
    """Updates a line with the specified key from the specified src dict.

    Arguments:
        lines (list): Existing list of lines.
        src (dict): Source dict from which to update the value.
        empty_values (list): Values that should result in removing the line
            instead of updating it.
        ini_key (str): Key in lines to update (eg 'ListenPort').
        src_keys (list): Src keys to check (eg ['listen_port', 'port']).
    """
    if not src_keys:
        src_keys = [underscore(ini_key)]

    value = None
    for x in src_keys:
        value = src.get(x)
        if value is not None:
            break

    if value in empty_values:
        replace_ini_line_value(lines, ini_key, None)
    elif value is True:
        replace_ini_line_value(lines, ini_key, "true")
    elif value:
        replace_ini_line_value(lines, ini_key, value)


def _debug_wg_cnf_sections(sections):
    """Logs content of wg cnf ini.

    Arguments:
        sections (list): List of list of lines.
    """
    log = getLogger(__name__)
    if log.level <= DEBUG:
        log.debug("\n".join(_debug_wg_cnf_sections_content(sections)))


# simpler for now to keep this logic together rather than subdivide it more functions
def _debug_wg_cnf_sections_content(sections):  # noqa: CCR001
    """Generates content lines from wg cnf ini to log.

    Arguments:
        sections (list): List of list of lines.

    Returns:
        list: List of of lines to log.
    """
    content = [""]
    for lines in sections:
        for line in lines:
            if line.startswith("PrivateKey"):
                line = _hide_private_key_line(line)
            elif line.startswith("PresharedKey"):
                line = _hide_preshared_key_line(line)
            content.append(line)
    return content


def _hide_private_key_line(line):
    """Hides the private key on the specified line.

    Arguments:
        line (str): WG conf line (eg 'PrivateKey = XYZ=').

    Returns:
        str: Line with hidden key ('PrivateKey = (hidden; public key: ABC=)').
    """
    parts = line.split("=", maxsplit=1)
    if len(parts) < 2:
        return line

    name = parts[0].strip()
    value = derive_public_key(parts[1].strip())
    return f"{name} = (hidden; public key: {value})"


def _hide_preshared_key_line(line):
    """Hides the preshared key on the specified line.

    Arguments:
        line (str): WG conf line (eg 'PresharedKey = XYZ=').

    Returns:
        str: Line with hidden key ('PresharedKey = (hidden; sha256 hash: ABC=)').
    """
    parts = line.split("=", maxsplit=1)
    if len(parts) < 2:
        return line

    name = parts[0].strip()
    value = hash_preshared_key(parts[1].strip())
    return f"{name} = (hidden; sha256 hash: {value})"



# -*- coding: utf-8 -*-
"""Utilities for updating endpoints with resolved hostnames."""

from base64 import urlsafe_b64encode
from contextlib import suppress
from logging import getLogger
from re import fullmatch, search
from socket import AF_INET, AF_INET6, SOCK_DGRAM, SOCK_STREAM, SOL_SOCKET, getaddrinfo
from socket import socket as make_socket
from time import time
from urllib.parse import urlsplit, urlunsplit

from dns.exception import DNSException
from dns.message import from_wire as message_from_wire
from dns.message import make_query
from dns.name import from_text as make_qname
from dns.query import BadResponse
from dns.query import _connect as connect_tcp
from dns.query import tcp as query_tcp
from dns.query import udp as query_udp
from dns.rcode import NOERROR, NXDOMAIN
from dns.rcode import to_text as format_rcode
from dns.rdataclass import IN
from dns.rdatatype import AAAA, ANY, A
from dns.resolver import Answer, Cache, NoAnswer
from requests import Request, Session
from requests.adapters import HTTPAdapter

from procustodibus_agent.cnf import DEFAULT_DNS_SERVERS
from procustodibus_agent.wg import run_wg_set

QUAD_SERVERS = {"1.1.1.1", "8.8.8.8", "9.9.9.9"}
PREFER_IPV6_SERVERS = DEFAULT_DNS_SERVERS[2:4] + DEFAULT_DNS_SERVERS[0:2]
# linux/include/uapi/asm-generic/socket.h
SO_MARK = 36


def is_likely_ip(hostname):
    """Checks if a hostname is likely an IP address.

    Arguments:
        hostname (str): Hostname to check (eg 'foo.example.com').

    Returns:
        True if likely an IP address.
    """
    return bool(
        hostname
        and (
            search(r"[0-9](\.[0-9]{1,3}){3}\]?$", hostname)
            or (
                fullmatch(r"\[?[0-9A-Fa-f:]+\]?", hostname)
                and search(":[^:]*:", hostname)
            )
        )
    )


def lookup_ip(cnf, hostname, family=None, raises=True):
    """Looks up the top IP address for the specified hostname.

    Arguments:
        cnf (Config): Config object.
        hostname (str): Hostname to lookup (eg 'foo.example.com').
        family (int): Preferred family (eg AF_INET6).
        raises (bool): True to raise if cannot resolve hostname (default True).

    Returns:
        str: Top IP address (eg '10.1.2.3').

    Propagates:
        DNSException: Cannot resolve hostname.
    """
    ips = lookup_ips(cnf, hostname, family, raises)
    return ips[0] if ips else ""


def lookup_ips(cnf, hostname, family=None, raises=True):
    """Looks up the IP addresses for the specified hostname.

    Arguments:
        cnf (Config): Config object.
        hostname (str): Hostname to lookup (eg 'foo.example.com').
        family (int): Preferred family (eg AF_INET6).
        raises (bool): True to raise if cannot resolve hostname (default True).

    Returns:
        list: IP addresss (eg ['10.1.2.3', '10.4.5.6']).

    Raises:
        Exception: Cannot resolve hostname.
    """
    try:
        return get_resolver(cnf).lookup_ips(hostname, family)
    except Exception as e:
        if raises:
            raise e
        message = f"dns error resolving {hostname}: {e}"
        getLogger(__name__).debug(message)
        return []


def get_resolver(cnf):
    """Finds or creates resolver for the specified configuration object.

    Arguments:
        cnf (Config): Config object.

    Returns:
        Resolver: Resolver object.
    """
    resolver = cnf.resolver
    if not resolver:
        resolver = DnsResolver(cnf)
        cnf.resolver = resolver
    return resolver


class Resolver(dict):
    """Base hostname resolver."""

    def lookup_ips(self, hostname, family=None):
        """Looks up the IP addresses for the specified hostname.

        Arguments:
            hostname (str): Hostname to lookup (eg 'foo.example.com').
            family (int): Preferred family (eg AF_INET6).

        Raises:
            DNSException: Cannot resolve hostname.
        """
        raise DNSException("cannot resolve hostnames")


class DictionaryResolver(Resolver):
    """Resolves hostnames through static dictionary."""

    def __init__(self, dictionary=None):
        """Creates new resolver object.

        Arguments:
            dictionary (dict): Map of hostnames to ip addresses.
        """
        self.dictionary = {} if dictionary is None else dictionary

    def __str__(self):  # noqa: D105
        return str(self.dictionary)

    def lookup_ips(self, hostname, family=None):
        """Looks up the IP addresses for the specified hostname.

        Arguments:
            hostname (str): Hostname to lookup (eg 'foo.example.com').
            family (int): Preferred family (eg AF_INET6).

        Returns:
            list: IP addresss (eg ['10.1.2.3', '10.4.5.6']).

        Raises:
            DNSException: Cannot resolve hostname.
        """
        try:
            return self.dictionary[hostname]
        except KeyError:
            raise DNSException(f"hostname not found: {hostname}")

    def __bool__(self):  # noqa: D105
        return bool(self.dictionary)


class DnsResolver(Resolver):
    """Resolves hostnames through DNS queries, caching results."""

    class _MinimumCnf(dict):
        """Configuration object with defaults for resolver."""

        def __init__(self, resolve_hostnames=""):
            """Creates minimal configuration object.

            Arguments:
                resolve_hostnames (str): Family preference (eg "ipv4" or "ipv6").
            """
            self.dns = []
            self.dns_protocol = ""
            self.doh = ""
            self.resolve_hostnames = resolve_hostnames
            self.socket_mark = ""
            self.transport = None

    def __init__(self, cnf=None, resolve_hostnames=""):
        """Creates new resolver object.

        Arguments:
            cnf (Config): Config object.
            resolve_hostnames (str): Family preference (eg "ipv4" or "ipv6").
        """
        self.cnf = cnf or self._MinimumCnf(resolve_hostnames)
        self.cache = Cache()
        self.prefer_ipv6, self.has_ipv6 = self._calculate_prefer_ipv6()
        self.dns_servers = self._calculate_dns_servers()
        self.doh_url, self.doh_hostname = self._calculate_doh_url()
        self.doh_sessions = {}

    def __str__(self):  # noqa: D105
        return f"{self.cnf.doh}#{self.cnf.dns}"

    def lookup_ips(self, hostname, family=None):
        """Looks up the IP addresses for the specified hostname.

        Arguments:
            hostname (str): Hostname to lookup (eg 'foo.example.com').
            family (int): Preferred family (eg AF_INET6).

        Returns:
            list: IP addresss (eg ['10.1.2.3', '10.4.5.6']).

        Propagates:
            DNSException: Cannot resolve hostname.
        """
        return self.lookup_ips_from_cache(hostname) or self.lookup_ips_from_servers(
            hostname, family
        )

    def lookup_ips_from_cache(self, hostname):
        """Looks up the IP addresses for the specified hostname in the cache.

        Arguments:
            hostname (str): Hostname to lookup (eg 'foo.example.com').

        Returns:
            list: IP addresses, or empty if not in cache.

        Raises:
            NoAnswer: Cached response was that the hostname has no IP addresses.
        """
        if not self.dns_servers:
            return []

        qname = make_qname(hostname)
        answer = self.cache.get((qname, ANY, IN))

        # NB: the Answer object is falsy if it has no rrset sections
        # (ie the server answered with an empty list of addresses, aka NXDOMAIN)
        if answer is None:
            return []
        if not answer:
            raise NoAnswer(response=answer.response)

        return [x.address for x in answer]

    # simpler to keep logic in same function even if it makes cognitive-complexity high
    def lookup_ips_from_servers(self, hostname, family=None):  # noqa: CCR001
        """Looks up the IP addresses for the specified hostname from DNS.

        Arguments:
            hostname (str): Hostname to lookup (eg 'foo.example.com').
            family (int): Preferred family (eg AF_INET6).

        Returns:
            list: IP addresss (eg ['10.1.2.3', '10.4.5.6']).

        Raises:
            NoAnswer: Hostname has no IP address.
            DNSException: Cannot resolve hostname.
        """
        servers = self.dns_servers

        if not servers:
            return self.lookup_ips_from_gai(hostname, family)

        if len(servers) > 1:
            for server in servers:
                try:
                    return self.lookup_ips_from_server(hostname, server, family)
                except NoAnswer as e:
                    raise e
                except Exception as e:
                    message = f"{server} dns error resolving {hostname}: {e}"
                    getLogger(__name__).debug(message)
                    self._adjust_dns_server_preference(server)

        try:
            return self.lookup_ips_from_server(hostname, servers[0], family, 5)
        except DNSException as e:
            raise e
        except Exception as e:
            raise DNSException(f"dns error resolving {hostname}: {e}") from e

    def lookup_ips_from_gai(self, hostname, family=None):
        """Looks up the IP addresses for the specified hostname from OS.

        Arguments:
            hostname (str): Hostname to lookup (eg 'foo.example.com').
            family (int): Preferred family (eg AF_INET6).

        Returns:
            list: IP addresss (eg ['10.1.2.3', '10.4.5.6']).

        Raises:
            DNSException: Cannot resolve hostname.
        """
        family = family or (AF_INET6 if self.prefer_ipv6 else AF_INET)
        try:
            return [x[4][0] for x in getaddrinfo(hostname, 0, family)]
        except Exception as e:
            family = AF_INET6 if family == AF_INET else AF_INET
            try:
                return [x[4][0] for x in getaddrinfo(hostname, 0, family)]
            except Exception:
                raise DNSException(f"dns error resolving {hostname}: {e}") from e

    def lookup_ips_from_server(self, hostname, server, family=None, timeout=0.5):
        """Looks up IP address for the specified hostname from the specified server.

        Arguments:
            hostname (str): Hostname to lookup (eg 'foo.example.com').
            server (str): DNS server to query (eg '9.9.9.9').
            family (int): Preferred family (eg AF_INET6).
            timeout (float): Timeout in seconds after which to error (default 0.5).

        Returns:
            list: IP addresss (eg ['10.1.2.3', '10.4.5.6']).

        Raises:
            NoAnswer: Hostname has no IP addresses.
            ErrorAnswer: Error querying the DNS server.
        """
        qname = make_qname(hostname)
        prefer_ipv6 = family == AF_INET6 if family else self.prefer_ipv6
        types = [AAAA, A] if prefer_ipv6 else [A, AAAA]

        (code, response, answer) = self.query_types_from_server(
            qname, server, types, timeout
        )
        if answer is not None:
            self.cache.put((qname, ANY, IN), answer)

        if code not in (NOERROR, NXDOMAIN):
            raise ErrorAnswer(response=response)
        if not answer:
            raise NoAnswer(response=response)

        return [x.address for x in answer]

    def query_types_from_server(self, qname, server, types, timeout=0.5):
        """Queries the specified server for the specified name of each specified type.

        Returns the response and answer for the first type with a non-empty answer;
        or the first response and answer if all types had an empty answer;
        or the first response and no answer if an error occurred querying the server.

        Arguments:
            qname (str): Fully-qualified name to lookup (eg 'foo.example.com.').
            server (str): DNS server to query (eg '9.9.9.9').
            types (list): List of record types to query (eg [A, AAAA]).
            timeout (float): Timeout in seconds after which to error (default 0.5).

        Returns:
            tuple: Response code, response object, answer object.
        """
        rdtype = types.pop(0)
        query = make_query(qname, rdtype)
        (response, port) = self.query_server(query, server, timeout)
        code = response.rcode()

        if code not in (NOERROR, NXDOMAIN):
            return (code, response, None)

        answer = Answer(qname, rdtype, IN, response, server, port)
        if not answer and len(types):
            try:
                (c, r, a) = self.query_types_from_server(qname, server, types, timeout)
                if a:
                    return (c, r, a)
            except Exception as e:
                message = f"{server} dns error resolving {query.question[0]}: {e}"
                getLogger(__name__).debug(message)

        return (code, response, answer)

    def query_server(self, query, server, timeout=0.5):
        """Sends the specified query to the specified server.

        Arguments:
            query (Message): Message object.
            server (str): DNS server to query (eg '9.9.9.9').
            timeout (float): Timeout in seconds after which to error (default 0.5).

        Returns:
            tuple: Response object, server port used.
        """
        protocol = self.cnf.dns_protocol
        if (not protocol or protocol == "https") and self.cnf.doh:
            return (self.query_with_doh(query, server, timeout), 443)
        if protocol == "tcp":
            return (self.query_with_tcp(query, server, timeout), 53)
        return (self.query_with_udp(query, server, timeout), 53)

    def query_with_udp(self, query, server, timeout=0.5):
        """Sends the specified query to the specified server with UDP.

        Arguments:
            query (Message): Message object.
            server (str): DNS server to query (eg '9.9.9.9').
            timeout (float): Timeout in seconds after which to error (default 0.5).

        Returns:
            Response object.
        """
        socket = self._make_socket(server, SOCK_DGRAM)
        try:
            return query_udp(query, server, timeout, sock=socket)
        finally:
            socket.close()

    def query_with_tcp(self, query, server, timeout=0.5):
        """Sends the specified query to the specified server with TCP.

        Arguments:
            query (Message): Message object.
            server (str): DNS server to query (eg '9.9.9.9').
            timeout (float): Timeout in seconds after which to error (default 0.5).

        Returns:
            Response object.
        """
        socket = self._make_socket(server, SOCK_STREAM)
        try:
            connect_tcp(socket, (server, 53), time() + timeout)
            return query_tcp(query, server, timeout, sock=socket)
        finally:
            socket.close()

    def query_with_doh(self, query, server, timeout=0.5):
        """Sends the specified query to the specified server with DoH.

        Arguments:
            query (Message): Message object.
            server (str): DNS server to query (eg '9.9.9.9').
            timeout (float): Timeout in seconds after which to error (default 0.5).

        Returns:
            Response object.
        """
        session = self._get_session(server)
        return query_doh(query, self.doh_url, timeout, session)

    def __bool__(self):  # noqa: D105
        return True

    def _make_socket(self, ip, socket_type):
        """Creates a new socket of the specified type.

        Arguments:
            ip (str): Server IP address.
            socket_type (int): SOCK_STREAM or SOCK_DGRAM.

        Raises:
            Exception: Socket set-up failed.

        Returns:
            Socket object.
        """
        socket = make_socket(AF_INET6 if ":" in ip else AF_INET, socket_type)
        try:
            socket.setblocking(False)
            mark = coerce_socket_mark(self.cnf.socket_mark)
            if mark:
                socket.setsockopt(SOL_SOCKET, SO_MARK, mark)
            return socket
        except Exception:
            socket.close()
            raise

    def _get_session(self, ip):
        """Finds or creates a new requests session for the specified server IP.

        Arguments:
            ip (str): Server IP address.

        Returns:
            Session object.
        """
        session = self.doh_sessions.get(ip)
        if not session:
            session = self._make_session(ip)
            self.doh_sessions[ip] = session
        return session

    def _make_session(self, ip):
        """Creates a new requests session for the specified server IP.

        Arguments:
            ip (str): Server IP address.

        Returns:
            Session object.
        """
        resolver = DictionaryResolver(dictionary={self.doh_hostname: [ip]})
        adapter = ResolverHTTPAdapter(
            resolver=resolver,
            socket_mark=self.cnf.socket_mark,
        )
        session = Session()
        session.mount("http://", adapter)
        session.mount("https://", adapter)
        return session

    def _adjust_dns_server_preference(self, server):
        """Adjust DNS preferences based on failed connection to the specified server.

        Arguments:
            server (str): Server IP address.
        """
        if (
            self.prefer_ipv6
            and self.cnf.resolve_hostnames != "ipv6"
            and server == DEFAULT_DNS_SERVERS[-1]
            and self.dns_servers == PREFER_IPV6_SERVERS
        ):
            getLogger(__name__).info("adjusting preference from ipv6 to ipv4")
            self.prefer_ipv6 = False
            self.dns_servers = DEFAULT_DNS_SERVERS
            self.cache.flush()
        elif (
            self.has_ipv6
            and not self.prefer_ipv6
            and self.cnf.resolve_hostnames != "ipv4"
            and server == PREFER_IPV6_SERVERS[-1]
            and self.dns_servers == DEFAULT_DNS_SERVERS
        ):
            getLogger(__name__).info("adjusting preference from ipv4 to ipv6")
            self.prefer_ipv6 = True
            self.dns_servers = PREFER_IPV6_SERVERS
            self.cache.flush()

    def _calculate_prefer_ipv6(self):
        """Checks if IPv6 is preferred and available.

        Returns:
            tuple: True if IPv6 preferred, True if IPv6 available
        """
        if self.cnf.resolve_hostnames == "ipv6":
            return (True, True)
        if self.cnf.resolve_hostnames == "ipv4":
            return (False, False)
        # if no explicit preference, use ipv6 if preferred address for localhost is ipv6
        try:
            families = [x[0] for x in getaddrinfo("localhost", 0)]
            return (families[0] == AF_INET6, AF_INET6 in families)
        except Exception:
            return (False, False)

    def _calculate_dns_servers(self):
        """Returns list of DNS servers in preferred order.

        Returns:
            list: List of DNS servers.
        """
        servers = self.cnf.dns
        servers_6 = [x for x in servers if ":" in x]
        servers_4 = [x for x in servers if ":" not in x]

        # if ipv6 not available, use just ipv4 servers
        if not self.has_ipv6 and servers_4:
            return servers_4
        # when ipv6 is preferred, reorder default ipv6 servers first
        elif self.prefer_ipv6 and len(servers) > 1 and servers[0] in QUAD_SERVERS:
            return servers_6 + servers_4
        return servers

    def _calculate_doh_url(self):
        """Returns URL and hostname to use for DoH.

        Returns:
            Tuple of DoH URL, DoH hostname
            (eg ('https://example.org/dns-query', 'example.org')).
        """
        doh = self.cnf.doh
        if doh.startswith("https:") or doh.startswith("http:"):
            return (doh, urlsplit(doh).hostname)
        return (f"https://{doh}:443/dns-query", doh)


class ErrorAnswer(NoAnswer):
    """The DNS response was an error."""

    supp_kwargs = {"response"}

    def __init__(self, *args, **kwargs):  # noqa: D107
        super().__init__(*args, **kwargs)

    def __str__(self):  # noqa: D105
        return f"The DNS response was {self.rcode()} to the question: {self.query()}"

    def query(self):
        """Formats DNS query.

        Returns:
            str: DNS query.
        """
        q = self.kwargs["response"].question
        return str(q[0]) if q else ""

    def rcode(self):
        """Formats DNS response code.

        Returns:
            str: DNS response code.
        """
        return format_rcode(self.kwargs["response"].rcode())


class ResolverHTTPAdapter(HTTPAdapter):
    """Custom requests library HTTP adapter."""

    def __init__(self, *args, **kwargs):
        """Creates new adapter.

        Arguments:
            *args (list): List of arguments to pass to super.
            *kwargs (dict): Dictionary of arguments to pass to super.
        """
        self.resolver = kwargs.pop("resolver", None)
        self.socket_mark = kwargs.pop("socket_mark", None)
        super(ResolverHTTPAdapter, self).__init__(*args, **kwargs)

    def init_poolmanager(self, *args, **kwargs):
        """Initializes the adapter's connection pool manager.

        Arguments:
            *args (list): List of arguments to pass to super.
            *kwargs (dict): Dictionary of arguments to pass to super.
        """
        mark = coerce_socket_mark(self.socket_mark)
        if mark:
            kwargs["socket_options"] = [(SOL_SOCKET, SO_MARK, mark)]
        super(ResolverHTTPAdapter, self).init_poolmanager(*args, **kwargs)

    def send(self, request, **kwargs):
        """Sends the specified request.

        Arguments:
            request: Request object.
            *kwargs (dict): Dictionary of arguments to pass to super.

        Returns:
            Response object.
        """
        url = urlsplit(request.url)
        if not is_likely_ip(url.hostname):
            self._replace_hostname_with_ip(request, url)

        return super(ResolverHTTPAdapter, self).send(request, **kwargs)

    def _replace_hostname_with_ip(self, request, url):
        """Replaces hostname in request URL with resolved IP address.

        Arguments:
            request: Request object.
            url: Parsed URL object.
        """
        hostname = url.hostname
        ip = self.resolver.lookup_ips(hostname)[0] if self.resolver else hostname
        x = format_endpoint_address(ip, url.port or 443)

        request.url = urlunsplit((url.scheme, x, url.path, url.query, url.fragment))
        request.headers["host"] = hostname

        pool_kw = self.poolmanager.connection_pool_kw
        pool_kw["assert_hostname"] = hostname
        pool_kw["server_hostname"] = hostname


def query_doh(query, url, timeout, session):
    """Queries the specified DoH URL with the specified DNS query.

    Arguments:
        query: DNS query object.
        url (str): Base DoH URL (eg 'https://example.org/dns-query').
        timeout (float): Request timeout in seconds (eg '0.5').
        session: Requests session object.

    Returns:
        DNS response object.

    Raises:
        BadResponse: Bad response from DoH URL.

    Propagates:
        RequestException: Error connecting to DoH URL.
    """
    http = _send_doh_query(query, url, timeout, session)
    response = message_from_wire(http.content)
    if not query.is_response(response):
        raise BadResponse
    return response


def _send_doh_query(query, url, timeout, session):
    """Sends the specified DNS query to the specified DoH URL.

    Arguments:
        query: DNS query object.
        url (str): Base DoH URL (eg 'https://example.org/dns-query').
        timeout (float): Request timeout in seconds (eg '0.5').
        session: Requests session object.

    Returns:
        Requests response object.

    Propagates:
        RequestException: Error connecting to DoH URL.
    """
    params = {"dns": urlsafe_b64encode(query.to_wire()).decode("utf-8").rstrip("=")}
    headers = {"accept": "application/dns-message"}
    request = Request("GET", url, params=params, headers=headers)
    request = request.prepare()
    response = session.send(request, timeout=timeout)
    response.raise_for_status()
    return response


def coerce_socket_mark(mark):
    """Converts the socket mark to an int.

    Arguments:
        mark: Mark value (eg '' or 'off' or 123 or '0x7b').

    Returns:
        int: Mark value (eg 0 or 123).
    """
    if not mark or mark == "off":
        return 0
    if isinstance(mark, int):
        return mark
    return int(str(mark), 0)


def resolve_endpoint_hostname(cnf, endpoint):
    """Resolves the specified endpoint to an IP address.

    Arguments:
        cnf (Cnf): Config object.
        endpoint (str): Endpoint with a hostname (eg 'vpn.example.com:51820').

    Returns:
        tuple: Resolved endpoint, hostname, port
        (eg ('[fd01::1]:51820', 'vpn.example.com', 51820)).
    """
    hostname, port, family = split_endpoint_address(endpoint)
    if not hostname or family:
        return endpoint, "", 0

    ip = lookup_ip(cnf, hostname, raises=False)
    if ip:
        endpoint = format_endpoint_address(ip, port)
    return endpoint, hostname, port


def apply_endpoint_hostnames(cnf, interfaces):
    """Updates the endpoints of the specified peers if configured with a hostname.

    Arguments:
        cnf (Cnf): Config object.
        interfaces (dict): Interfaces data.
    """
    if cnf.resolve_hostnames == "once":
        return

    for name, interface in interfaces.items():
        for pubkey, peer in interface.get("peers", {}).items():
            hostname = peer.get("hostname")
            endpoint = peer.get("endpoint")
            apply_endpoint_hostname(cnf, name, pubkey, hostname, endpoint)


def apply_endpoint_hostname(cnf, name, pubkey, hostname, endpoint):
    """Updates the specified peer's endpoint using the specified hostname.

    Arguments:
        cnf (Cnf): Config object.
        name (str): Interface name (eg 'wg0').
        pubkey (str): Peer public key.
        hostname (str): Hostname to use (eg 'foo.example.com').
        endpoint (str): Existing endpoint (eg '10.10.10.10:51820').
    """
    if not name or not pubkey or not hostname or not endpoint:
        return

    ip, port, family = split_endpoint_address(endpoint)
    ips = lookup_ips(cnf, hostname, family, raises=False)
    if ips and ip not in ips:
        set_endpoint_address(cnf, name, pubkey, ips[0], port)


def set_endpoint_address(cnf, name, pubkey, address, port):
    """Runs the `wg set` command to set the endpoint of the specified peer.

    Arguments:
        cnf (Cnf): Config object.
        name (str): Interface name (eg 'wg0').
        pubkey (str): Peer public key.
        address (str): Endpoint IP address (eg '10.10.10.10').
        port (int): Endpoint port (eg 51820).
    """
    endpoint = format_endpoint_address(address, port)
    run_wg_set(cnf, [name, "peer", pubkey, "endpoint", endpoint])


def format_endpoint_address(address, port):
    """Formats the specified endpoint IP address and port.

    Arguments:
        address (str): Endpoint IP address (eg 'fd01::1').
        port (int): Endpoint port (eg 51820).

    Returns:
        str: Formatted endpoint address (eg '[fd01::1]:51820').
    """
    port = port or 51820
    return f"[{address}]:{port}" if address.find(":") >= 0 else f"{address}:{port}"


def split_endpoint_address(endpoint):
    """Splits the specified endpoint into IP address, port, and address family.

    Arguments:
        endpoint (str): IP address and port (eg '[fc00:0:0:1::]:51820').

    Returns:
        tuple: String IP address, integer port number, integer family.
    """
    if not endpoint:
        return "", 0, 0

    ipv6 = fullmatch(r"\[([^\]]+)\](?::(\d+))?", endpoint)
    if ipv6:
        return ipv6[1], int(ipv6[2] or 0), AF_INET6

    family = 0 if search("[A-Za-z]", endpoint) else AF_INET
    ipv4 = endpoint.split(":")
    if len(ipv4) == 2:
        with suppress(ValueError):
            return ipv4[0], int(ipv4[1]), family

    return ipv4[0], 0, family


def split_endpoint_hostname_and_port(endpoint):
    """Splits the specified endpoint into hostname and port.

    If the endpoint has an IP address instead of a hostname, returns ('', 0).

    Arguments:
        endpoint (str): Hostname and port (eg 'vpn.example.com:51820').

    Returns:
        Tuple of string hostname and integer port number.
    """
    address, port, family = split_endpoint_address(endpoint)
    if family:
        return "", 0
    return address, port



# -*- coding: utf-8 -*-
"""WireGuard utilities."""

import re
from base64 import b64decode, b64encode
from hashlib import sha256
from pathlib import Path
from platform import system
from subprocess import PIPE, run  # noqa: S404

from nacl.encoding import Base64Encoder
from nacl.public import PrivateKey


def get_wg_version(cnf):
    """Returns WireGuard client version.

    Arguments:
        cnf (Config): Config object.

    Returns:
        str: Version string (eg '1.2.3').
    """
    if system() == "Windows":
        return _get_windows_wg_version(cnf)
    else:
        return _get_standard_wg_version(cnf)


def _get_standard_wg_version(cnf):
    """Returns standard WireGuard client version.

    Arguments:
        cnf (Config): Config object.

    Returns:
        str: Version string (eg '1.2.3').
    """
    result = run([cnf.wg, "--version"], stdout=PIPE)  # noqa: S603 S607
    output = result.stdout.decode("utf-8")
    return re.search(r"\sv(\S+)", output)[1]


def _get_windows_wg_version(cnf):
    """Returns Windows WireGuard client version.

    Arguments:
        cnf (Config): Config object.

    Returns:
        str: Version string (eg '1.2.3').
    """
    exe = _get_windows_wireguard_exe(cnf)
    escaped = re.sub(r"[/\\]", r"\\\\", str(exe))
    result = run(  # noqa: S603 S607
        ["wmic", "datafile", "where", f"name='{escaped}'", "get", "version", "/value"],
        stdout=PIPE,
    )
    output = result.stdout.decode("utf-8").strip()
    return re.sub(r"Version=", "", output)


def _get_windows_wireguard_exe(cnf):
    """Returns path to Windows WireGuard client.

    Arguments:
        cnf (Config): Config object.

    Returns:
        Path: Path to WireGuard executable.
    """
    wg = Path(cnf.wg)
    if cnf.wiresock:
        return wg
    directory = wg.parent
    if not directory.name:
        directory = Path("C:/Program Files/WireGuard/")
    return Path(directory, "wireguard.exe")


def run_wg_set(cnf, args):
    """Runs `wg set` command.

    Arguments:
        cnf (Config): Config object.
        args: Command arguments.
    """
    if not cnf.wiresock:
        run([cnf.wg, "set", *args])  # noqa: S603 S607


def run_wg_show(cnf):
    """Runs `wg show` command.

    Arguments:
        cnf (Config): Config object.

    Returns:
        str: Output of `wg show` command.
    """
    if cnf.wiresock:
        return ""
    result = run([cnf.wg, "show", "all", "dump"], stdout=PIPE)  # noqa: S603 S607
    return result.stdout.decode("utf-8")


def parse_wg_show(s):
    """Parses `wg show` command output into a dict.

    Raises:
        ValueError: Output can't be parsed.

    Arguments:
        s (str): Output of `wg show` command.

    Returns:
        dict: Output of `wg show` command as a dict.
    """
    result = {}

    for line in s.splitlines():
        fields = line.split()
        len_fields = len(fields)

        if len_fields == 0:
            pass
        elif len_fields == 5:
            _add_interface(result, fields)
        elif len_fields == 9:
            _add_peer(result, fields)
        else:
            raise ValueError("can't parse line from wg show", line)

    return result


def update_socket_mark(data, cnf):
    """Updates the active socket mark with a new mark from the parsed interfaces.

    Arguments:
        data (dict): Dict parsed from `wg show` command.
        cnf (Config): Config object.

    Returns:
        dict: Same dict.
    """
    # don't override explicitly configured mark
    if cnf.fw_mark:
        return data
    fwmark = _get_fwmark_for_default_peer(data)
    # reset mark consumers if changed
    if cnf.socket_mark != fwmark:
        cnf.socket_mark = fwmark
        cnf.resolver = None
        cnf.transport = None
    return data


def filter_wg_show(data, cnf):
    """Removes configured elements from parsed output of `wg show`.

    Arguments:
        data (dict): Dict parsed from `wg show` command.
        cnf (Config): Config object.

    Returns:
        dict: Same dict with some items removed.
    """
    for interface in cnf.unmanaged_interfaces:
        data.pop(interface, None)
    if cnf.redact_secrets:
        data = remove_secrets_from_wg_show(data)
    return data


def remove_secrets_from_wg_show(data):
    """Removes secrets from parsed output of `wg show`.

    Arguments:
        data (dict): Dict parsed from `wg show` command.

    Returns:
        dict: Same dict with secret items removed.
    """
    for interface in data.values():
        del interface["private_key"]
        for peer in interface["peers"].values():
            del peer["preshared_key"]
    return data


def _add_interface(result, fields):
    """Adds the the specified interface fields to the result.

    Arguments:
        result (dict): Result dict.
        fields (list): List of field values.
    """
    interface, private, public, port, fwmark = fields
    result[interface] = {
        "private_key": private if private != "(none)" else "",
        "public_key": public if public != "(none)" else "",
        "listen_port": int(port),
        "fwmark": int(fwmark, 0) if fwmark != "off" else 0,
        "peers": {},
    }


def _add_peer(result, fields):
    """Adds the the specified peer fields to the result.

    Arguments:
        result (dict): Result dict.
        fields (list): List of field values.
    """
    interface, public, preshared, endpoint, ips, handshake, rx, tx, keepalive = fields
    result[interface]["peers"][public] = {
        "preshared_key": preshared if preshared != "(none)" else "",
        "preshared_key_hash": hash_preshared_key(preshared),
        "endpoint": endpoint if endpoint != "(none)" else "",
        "allowed_ips": ips.split(",") if ips != "(none)" else [],
        "latest_handshake": int(handshake),
        "transfer_rx": int(rx),
        "transfer_tx": int(tx),
        "persistent_keepalive": int(keepalive) if keepalive != "off" else 0,
    }


def derive_public_key(key):
    """Derives the public key from the specified private key.

    Arguments:
        key (str): Private key as a base64 string.

    Returns:
        str: Public key as a base64 string.
    """
    if not key or key == "(none)":
        return ""
    public_key = PrivateKey(key, Base64Encoder).public_key
    return public_key.encode(Base64Encoder).decode("utf-8")


def hash_preshared_key(key):
    """Hashes the specified key.

    Arguments:
        key (str): Base64-encoded key to hash.

    Returns:
        str: Hash of raw key bytes as base64 string.
    """
    if not key or key == "(none)":
        return ""
    return b64encode(sha256(b64decode(key.encode())).digest()).decode()


# simpler to keep this logic together rather than subdivide it more functions
def separate_dns_and_search(items):  # noqa: CCR001
    """Separates the specified list into DNS servers and search domains.

    Arguments:
        items (list): Mixed list of DNS servers and search domains.

    Returns:
        tuple: List of DNS servers, list of search domains.
    """
    dns = []
    search = []
    if items:
        for item in items:
            if re.search("[A-Za-z]", item) and ":" not in item:
                search.append(item)
            else:
                dns.append(item)
    return dns, search


# simpler to keep this logic together rather than subdivide it more functions
def _get_fwmark_for_default_peer(data):  # noqa: CCR001
    """Returns the fwmark for the first interface with peer with a default route.

    Arguments:
        data (dict): Dict parsed from `wg show` command.

    Returns:
        str: Socket mark or blank.
    """
    for name in sorted(data):
        interface = data[name]
        fwmark = interface["fwmark"]
        if fwmark:
            for peer in interface["peers"].values():
                for ip in peer["allowed_ips"]:
                    if ip.endswith("/0"):
                        return fwmark
    return ""



# -*- coding: utf-8 -*-
"""Connectivity test."""
import os
import sys
from re import search

from dns.exception import DNSException
from requests import RequestException

from procustodibus_agent import DOCS_URL
from procustodibus_agent.api import (
    get_health_info,
    get_host_info,
    raise_unless_has_cnf,
    setup_api,
)
from procustodibus_agent.resolve_hostname import is_likely_ip, lookup_ip
from procustodibus_agent.wg import parse_wg_show, run_wg_show, update_socket_mark
from procustodibus_agent.wg_cnf import load_all_from_wg_cnf


def check_connectivity(cnf, output=None):
    """Runs all connectivity checks and outputs issues.

    Arguments:
        cnf (Config): Config object.
        output (IOBase): Output stream to write issues to (defaults to stdout).

    Returns:
        int: 0 if no issues, positive number if issues.
    """
    if not output:
        output = sys.stdout

    try:
        raise_unless_has_cnf(cnf)
    except Exception as e:
        # print to stdout or designated file
        print(str(e), file=output)  # noqa: T201
        return 1

    exit_code = (
        check_wg(cnf, output)
        + check_dns(cnf, output)
        + check_health(cnf, output)
        + check_host(cnf, output)
    )

    if exit_code:
        # print to stdout or designated file
        print(  # noqa: T201
            f"Issues encountered; see {DOCS_URL}/guide/agents/troubleshoot/ to fix",
            file=output,
        )
    else:
        # print to stdout or designated file
        print("All systems go :)", file=output)  # noqa: T201

    return exit_code


# simpler to keep this logic together rather than subdivide it more functions
def check_wg(cnf, output):  # noqa: CCR001
    """Checks that wireguard is available and configured with at least one interface.

    Arguments:
        cnf (Config): Config object.
        output (IOBase): Output stream to write issues to.

    Returns:
        int: 0 if no issues, positive number if issues.
    """
    if cnf.wiresock:
        try:
            interfaces = load_all_from_wg_cnf(cnf)
        except Exception as e:
            _bad(f"cannot open interface conf files ({e})", output)
            return 2
    else:
        try:
            interfaces = parse_wg_show(run_wg_show(cnf))
            update_socket_mark(interfaces, cnf)
        except OSError as e:
            _bad(f"no wg executable found ({e})", output)
            return 2

    if interfaces:
        _good(f"{len(interfaces)} wireguard interfaces found", output)
        return 0
    else:
        _bad("no wireguard interfaces found", output)
        return 0


def check_dns(cnf, output):
    """Checks that the local DNS resolver can resolve api.procustodib.us.

    Arguments:
        cnf (Config): Config object.
        output (IOBase): Output stream to write issues to.

    Returns:
        int: 0 if no issues, positive number if issues.
    """
    hostname = _get_hostname(cnf.api)
    try:
        address = hostname if is_likely_ip(hostname) else lookup_ip(cnf, hostname)
        _good(f"{address} is pro custodibus ip address", output)
        return 0
    except DNSException as e:
        _bad(e, output)
        return 4


def check_health(cnf, output):
    """Checks connectivity to and the health of the Pro Custodibus API.

    Arguments:
        cnf (Config): Config object.
        output (IOBase): Output stream to write issues to.

    Returns:
        int: 0 if no issues, positive number if issues.
    """
    try:
        errors = [x["error"] for x in get_health_info(cnf) if not x["healthy"]]
    except (DNSException, RequestException) as e:
        errors = [f"server unavailable ({e})"]

    if errors:
        for error in errors:
            _bad(f"unhealthy pro custodibus api: {error}", output)
        return 8
    else:
        _good("healthy pro custodibus api", output)
        return 0


def check_host(cnf, output):
    """Checks that the agent can access the configured host through the API.

    Arguments:
        cnf (Config): Config object.
        output (IOBase): Output stream to write issues to.

    Returns:
        int: 0 if no issues, positive number if issues.
    """
    try:
        _setup_if_available(cnf)
    except (DNSException, RequestException, ValueError) as e:
        _bad(f"cannot set up access to api ({e})", output)
        return 16

    try:
        host = get_host_info(cnf)
        name = host["data"][0]["attributes"]["name"]
        _good(f"can access host record on api for {name}", output)
        return 0
    except (DNSException, RequestException, ValueError) as e:
        _bad(f"cannot access host record on api ({e})", output)
        return 16


def _setup_if_available(cnf):
    """Sets up new agent credentials if setup code is available.

    Arguments:
        cnf (Config): Config object.
    """
    if type(cnf.setup) is dict or os.path.exists(cnf.setup):
        setup_api(cnf)


def _good(message, output):
    """Prints the specified "good" message to the specified output stream.

    Arguments:
        message (str): Message to print.
        output (IOBase): Output stream to write to.
    """
    # print to stdout or designated file
    print(f"... {message} ...", file=output)  # noqa: T201


def _bad(message, output):
    """Prints the specified "bad" message to the specified output stream.

    Arguments:
        message (str): Message to print.
        output (IOBase): Output stream to write to.
    """
    # print to stdout or designated file
    print(f"!!! {message} !!!", file=output)  # noqa: T201


def _get_hostname(url):
    """Extracts the hostname from the specified URL.

    Arguments:
        url (str): URL (eg 'http://test.example.com:8080').

    Returns:
        str: Hostname (eg 'test.example.com').
    """
    match = search(r"(?<=://)[^:/]+", url)
    return match.group(0) if match else None
